#include <stdio.h>
#include "util.h"
#include "vpr_types.h"
#include "globals.h"
#include "rr_graph2.h"
#include "rr_graph3.h"
#include "rr_graph_sbox.h"
#include <assert.h>


/* Variables below are the global variables shared only amongst the rr_graph *
 ************************************ routines. ******************************/

/* [0..num_rr_nodes-1].  TRUE if a node is already listed in the edges array *
 * that's being constructed.  This allows me to ensure that there are never  *
 *  duplicate edges (two edges between the same thing).                      */

boolean *rr_edge_done; 


/* Used to keep my own list of free linked integers, for speed reasons.     */

t_linked_edge *free_edge_list_head = NULL;



/*************************** Variables local to this module *****************/

/* Two arrays below give the rr_node_index of the channel segment at        *
 * (i,j,track) for fast index lookup.                                       */

static int ***chanx_rr_indices;  /* [1..nx][0..ny][0..nodes_per_chan-1] */
static int ***chany_rr_indices;  /* [0..nx][1..ny][0..nodes_per_chan-1] */
  


/************************** Subroutines local to this module ****************/

static int load_chanx_rr_indices (t_seg_details *seg_details_x, int 
           nodes_per_chan, int start_index, int i, int j); 

static int load_chany_rr_indices (t_seg_details *seg_details_y, int 
           nodes_per_chan, int start_index, int i, int j); 

/*---
static void get_switch_type (boolean is_from_sbox, boolean is_to_sbox, 
         short from_node_switch, short to_node_switch, short switch_types[2]); 
	 ---*/

/******************** Subroutine definitions *******************************/
// ======================================================================= //

t_seg_details *alloc_and_load_seg_details (int nodes_per_chan, t_segment_inf 
              *segment_inf, int num_seg_types, int max_len) {
 
/* Allocates and loads the seg_details data structure.  Max_len gives the   *
 * maximum length of a segment (dimension of array).  The code below tries  *
 * to:                                                                      *
 * (1) stagger the start points of segments of the same type evenly;        *
 * (2) spread out the limited number of connection boxes or switch boxes    *
 *     evenly along the length of a segment, starting at the segment ends;  *
 * (3) stagger the connection and switch boxes on different long lines,     *
 *     as they will not be staggered by different segment start points.     */
 
 int tracks_left, i, next_track, ntracks, itrack, length, j, index;
 int wire_switch, opin_switch, num_cb, num_sb;
 float frac_left, start_step, sb_step, cb_step;
 float cb_off_step, sb_off_step;
 t_seg_details *seg_details;
 boolean longline;

 seg_details = (t_seg_details *) my_malloc (nodes_per_chan * sizeof 
                    (t_seg_details));

 tracks_left = nodes_per_chan;
 next_track = 0;
 frac_left = 1.;

 // -------------------------------------------------------------------------
 // num_seg_types = 4 for Virtex like mixed architecture
 // segment types = {6, 2, 1, long-line}
 for (i=0;i<num_seg_types;i++) {
    if (frac_left <= 0.) {
       ntracks = 0;
    }
    else {
       ntracks = nint (segment_inf[i].frequency / frac_left * tracks_left);
       // E.g.: 7 = nint(6.6) = [ 0.6/1 ]*11 for i=2 for segment length 6
       
    }
    frac_left -= segment_inf[i].frequency;// 1-0.6 = 0.4
    tracks_left -= ntracks; // 11-7 = 4

    if (ntracks == 0) 
       continue;      /* No tracks of this length.  Avoid divide by 0, etc. */

    longline = segment_inf[i].longline;
 
    if (!longline) {  
       length = min (segment_inf[i].length, max_len);
       start_step = (float) length / (float) ntracks;
    }
    else {                                 /* Is a long line. */
       length = max_len;
       start_step = 0.;
    }

    // Number of connection blocks along this routing segment.
    // E.g.: there is only one CB on a length 1 routing-segment. 
    num_cb = nint (length * segment_inf[i].frac_cb);
    // Number of switch blocks(SB) along this routing segment.
    num_sb = nint ((length + 1.) * segment_inf[i].frac_sb);


/* Distribute connection boxes, with two endpoints covered first.           */
//------------------------ Spread CB's
    if (!longline) {
       if (num_cb > 1) 
          cb_step = (float) (length-1) / (float) (num_cb - 1);
       else
          cb_step = 0.;
    }

/* Treat longlines differently.  Distribute C and S blocks as if the        *
 * longlines were circular (i.e., don't cover both endpoints first).  This  *
 * is important, as the rotation of the C and S blocks from line to line    *
 * would cause stupidities otherwise.                                       */

    else {   
       if (num_cb > 0)
          cb_step = (float) length / (float) num_cb;
       else
          cb_step = 0.;
    }


/* Now spread out switch boxes in the same way.     */
//------------------------ Spread SB's

    if (!longline) {
       if (num_sb > 1) 
          sb_step = (float) length / (float) (num_sb - 1);
       else
          sb_step = 0.;
    }

    else {    /* Is a longline */
       if (num_sb > 0) 
          sb_step = (float) (length + 1) / (float) num_sb;
       else
          sb_step = 0;
    }

    if (!longline) {
       cb_off_step = 0.;
       sb_off_step = 0.;
    }

    else {                                 /* Is a long line. */
       cb_off_step = cb_step / (float) ntracks;
       sb_off_step = sb_step / (float) ntracks;
    }

    // segment_inf[i] a venit ca argument in aceasta functie.
    // wire_switch:  Index of the switch type that connects other wires *to* 
    //               this segment.  This is established thru .arch file
    //               for each type of segment.
    // opin_switch:  Index of the switch type that connects output pins (OPINs)
    //               *to* this segment.  
    wire_switch = segment_inf[i].wire_switch;
    opin_switch = segment_inf[i].opin_switch;
    // Aici ar trebui sa insert info about the fact a track is connected "to"
    // using a hard wire, based on the itrack.  Da dar aici construiesc seg_details
    // numai local adica ma uit la toate CW (ie nodes_per_chan) tracks of the channel
    // and go along all routing segments and establish where these segments have CB 
    // and SB's.  I cannot set here a fraction of all these SB's to be of type
    // hwrp_switch because I loose the global picture?  Rather I should do that when
    // I go and visit all (i,j) locations and construct rr_graph; there a certain 
    // percentage of all coneections inside all SB should be made T or L or + or -- or |.
    // These percentage shall apply to "all" connections involving segmentd of
    // length 1,2,6,long individually. 
    // For example, let's say on the overall chip i have 1000 connections (at SB level) 
    // beween four endings of four routing segments and 800 connections between 
    // middle (or along segments but different from endings). Also let's assume
    // that we want 50% (set by user) of all these connections to be hwrp's; that
    // is 500 + 400 connections.
    // Moreover, from routing analysis we know following usage fractions of routing
    // patterns (but notice that 400 connections can participate to build-up
    // patterns of only types T or L or + bu NOT -- or |):
    // +            10% ==> out of 500 + 400 connections 
    // 1T 2T 3T 4T  20% ...
    // 1L 2L 3L 4L  30% ...
    // --           20%
    // |            20%
    // oBs:  I will have to pay attention that as we "sweep" the fpga during
    // rr_graph construction we along channels x from left-->right and
    // along channels y from bottom-->top.  So pay attention to how
    //  1T 2T 3T 4T and 1L 2L 3L 4L are distributed during this "sweep".
    // 
    //         |   2
    //      --------- 
    //      |  |    | 3
    //  1T__|__|    |
    //      |  |    |
    //      ---------
    //         |   4
    //

    // Browseaza toate tracks cu acest tip de routing-segment.
    for (itrack=0;itrack<ntracks;itrack++) {
       seg_details[next_track].length = length;//next_track =1..nodes_per_chan
       seg_details[next_track].longline = longline;
 
       seg_details[next_track].start = nint (itrack * start_step) % length + 1;
       //printf(" s=%d",seg_details[next_track].start);
      // longline:  
      // start_step=0. --> .start=0 
      // all other segments, for example:  
      // start_step=1. --> .start=0, 1, etc.
      // si in felul asta as putea avea length 2 tracks starting from 
      // 0,0,1,1,and 2 totaling 5 tracks of this sort.
      // oBs:  Asta este stggering (un fel de ofset cu care incep tracks de 
      // un anumit fel incepind cu de la zero si urcind).
 
       seg_details[next_track].cb = (boolean *) my_calloc (length, 
                           sizeof (boolean));

       seg_details[next_track].sb = (boolean *) my_calloc ((length+1),
                            sizeof (boolean));
      // Mergind de unde incepe acest track-wire-segment si pina unde se
      // termina seteaza true or false daca are switch-box on it; apoi
      // connection-block.

       for (j=0;j<num_cb;j++) {
          index = nint (j * cb_step + itrack * cb_off_step) % length;
          seg_details[next_track].cb[index] = TRUE;
       }    

       for (j=0;j<num_sb;j++) {
          index = nint (j * sb_step + itrack * sb_off_step) % (length + 1);
          seg_details[next_track].sb[index] = TRUE;
       }    
 
       seg_details[next_track].Rmetal = segment_inf[i].Rmetal;
       seg_details[next_track].Cmetal = segment_inf[i].Cmetal;

       seg_details[next_track].wire_switch = wire_switch;
       seg_details[next_track].opin_switch = opin_switch;

       seg_details[next_track].index = i; // i=0,1,2,3 for {6,2,1,long-line}
                                          // next_track=1..nodes_per_chan (e.g. 11)


       next_track++;
    }// for (itrack=0;
 } // End for each segment type.
 // -------------------------------------------------------------------------

 
 assert (tracks_left == 0);
 assert (next_track == nodes_per_chan);
 return (seg_details);
}//eof
// ======================================================================= //
 
 
 
 
void free_seg_details (t_seg_details *seg_details, int nodes_per_chan) {
 
/* Frees all the memory allocated to an array of seg_details structures. */
 
 int i;
 
 for (i=0;i<nodes_per_chan;i++) {
    free (seg_details[i].cb);
    free (seg_details[i].sb);
 }
 free (seg_details);
}
// ======================================================================= //


void dump_seg_details (t_seg_details *seg_details, int nodes_per_chan, char 
           *fname) {

/* Dumps out an array of seg_details structures to file fname.  Used only   *
 * for debugging.                                                           */

 FILE *fp;
 int i, j;

 fp = my_fopen (fname, "w", 0);

 for (i=0;i<nodes_per_chan;i++) {
    fprintf (fp, "Track: %d.\n", i);
    fprintf (fp, "Length: %d,  Start: %d,  Long line: %d  wire_switch: %d  "
        "opin_switch: %d.\n", seg_details[i].length, seg_details[i].start, 
        seg_details[i].longline, seg_details[i].wire_switch, 
        seg_details[i].opin_switch);

    fprintf (fp, "Rmetal: %g  Cmetal: %g\n", seg_details[i].Rmetal, 
        seg_details[i].Cmetal);

    fprintf (fp, "cb list: ");
    for (j=0;j<seg_details[i].length;j++)
       fprintf (fp, "%d ", seg_details[i].cb[j]);

    fprintf (fp, "\nsb list: ");
    for (j=0;j<=seg_details[i].length;j++)
       fprintf (fp, "%d ", seg_details[i].sb[j]);

    fprintf (fp, "\n\n");
 } 

 fclose (fp);
}
// ======================================================================= //


int get_closest_seg_start (t_seg_details *seg_details, int itrack, int seg_num,
        int chan_num) {

/* Returns the segment number at which the segment this track lies on        *
 * started.                                                                  */

 int closest_start, length, start;

 if (seg_details[itrack].longline) {
    closest_start = 1;
 }
 else {
    length = seg_details[itrack].length;
    start = seg_details[itrack].start;

/* Start is guaranteed to be between 1 and length.  Hence adding length to *
 * the quantity in brackets below guarantees it will be nonnegative.       */

    closest_start = seg_num - (seg_num + chan_num - start + length) % length;
    closest_start = max (closest_start, 1);
 }

 return (closest_start);
}//eof
// ======================================================================= //



int get_clb_opin_connections (int ***clb_opin_to_tracks, int ipin, int i, int 
        j, int Fc_output, t_seg_details *seg_details_x, t_seg_details 
        *seg_details_y, t_linked_edge **edge_list_ptr, int nodes_per_chan, int 
        **rr_node_indices) {

/* Returns the number of tracks to which clb opin #ipin at (i,j) connects.   *
 * Also stores the nodes to which this pin connects in the linked list       *
 * pointed to by *edge_list_ptr.                                             */

 int iside, num_conn, iconn, itrack, tr_j, tr_i, to_node;
 t_rr_type to_rr_type;
 boolean cbox_exists;
 t_linked_edge *edge_list_head;
 t_seg_details *seg_details;

 edge_list_head = *edge_list_ptr;
 num_conn = 0;

 for (iside=0;iside<=3;iside++) {  
    if (clb_opin_to_tracks[ipin][iside][0] != OPEN) {
 
/* This track may be at a different (i,j) location than the clb.  Tracks *
 * above and to the right of a clb are at the same (i,j) -- the channels *
 * on the other two sides "belong" to different clbs.                    */
 
       if (iside == BOTTOM)
          tr_j = j-1;
       else
          tr_j = j;
 
       if (iside == LEFT)
          tr_i = i-1;
       else
          tr_i = i;
 
       if (iside == LEFT || iside == RIGHT) {
          to_rr_type = CHANY;
          seg_details = seg_details_y;
       }
       else {
          to_rr_type = CHANX;
          seg_details = seg_details_x;
       }
                 
       for (iconn=0;iconn<Fc_output;iconn++) {
          itrack = clb_opin_to_tracks[ipin][iside][iconn]; 
 
          if (to_rr_type == CHANX) 
             cbox_exists = is_cbox (tr_i, tr_j, itrack, seg_details);
          else
             cbox_exists = is_cbox (tr_j, tr_i, itrack, seg_details);

          if (cbox_exists) {
             to_node = get_rr_node_index (tr_i, tr_j, to_rr_type, itrack,
                           nodes_per_chan, rr_node_indices);
             edge_list_head = insert_in_edge_list (edge_list_head, to_node,
                       seg_details[itrack].opin_switch, &free_edge_list_head);
             num_conn++;
          }
       }
           
    }
 }

 if (num_conn == 0) {
    printf ("Error:  clb output pin %d at (%d,%d) does not connect to any "
            "tracks.\n", ipin, i, j);
    exit (1);
 }

 *edge_list_ptr = edge_list_head;
 return (num_conn);
}
// ======================================================================= //


int get_pad_opin_connections (int **pads_to_tracks, int ipad, int i, int j, 
        int Fc_pad, t_seg_details *seg_details_x, t_seg_details *seg_details_y, 
        t_linked_edge **edge_list_ptr, int nodes_per_chan, int 
        **rr_node_indices) {

/* Returns the number of tracks to which the pad opin at (i,j) connects.     *
 * If edge_ptr isn't NULL, it also loads the edge array.                     */

 int chan_i, chan_j, iconn, itrack, num_conn, to_node;
 t_rr_type chan_type;
 boolean cbox_exists;
 t_linked_edge *edge_list_head;
 t_seg_details *seg_details;

/* Find location of adjacent channel. */

 if (j == 0) {                 /* Bottom row of pads. */
    chan_i = i;
    chan_j = j;
    chan_type = CHANX;
    seg_details = seg_details_x;
 }
 else if (j == ny+1) {
    chan_i = i;
    chan_j = j-1;
    chan_type = CHANX;
    seg_details = seg_details_x;
 }
 else if (i == 0) {
    chan_i = i;
    chan_j = j;
    chan_type = CHANY;
    seg_details = seg_details_y;
 }
 else if (i == nx+1) {
    chan_i = i-1;
    chan_j = j;
    chan_type = CHANY;
    seg_details = seg_details_y;
 }
 else {
    printf ("Error in get_pad_opin_connections:  requested IO block at "
            "(%d,%d) does not exist.\n", i, j);
    exit (1);
 }

 edge_list_head = *edge_list_ptr;
 num_conn = 0;

 for (iconn=0;iconn<Fc_pad;iconn++) {
    itrack = pads_to_tracks[ipad][iconn];

    if (chan_type == CHANX)
       cbox_exists = is_cbox (chan_i, chan_j, itrack, seg_details);
    else
       cbox_exists = is_cbox (chan_j, chan_i, itrack, seg_details);

    if (cbox_exists) {
       to_node = get_rr_node_index (chan_i, chan_j, chan_type, itrack,
             nodes_per_chan, rr_node_indices);
       edge_list_head = insert_in_edge_list (edge_list_head, to_node, 
                       seg_details[itrack].opin_switch, &free_edge_list_head);
       num_conn++;
    }
 } 

 if (num_conn == 0) {
    printf ("Error:  INPAD %d at (%d,%d) does not connect to any "
            "tracks.\n", ipad, i, j);
    exit (1);
 }

 *edge_list_ptr = edge_list_head;
 return (num_conn);
}
// ======================================================================= //


boolean is_cbox (int seg_num, int chan_num, int itrack, t_seg_details 
       *seg_details) {

/* Returns 1 (TRUE) if the track segment at this segment and channel         *
 * location with track number itrack contains a connection box, 0 otherwise. */

 int seg_offset, start, length;

 length = seg_details[itrack].length;
 start = seg_details[itrack].start;

 seg_offset = (seg_num + chan_num - start + length) % length;
 return (seg_details[itrack].cb[seg_offset]);
}
// ======================================================================= //
// ======================================================================= //


int **alloc_and_load_rr_node_indices (int nodes_per_clb,
    int nodes_per_pad, int nodes_per_chan, t_seg_details *seg_details_x,
    t_seg_details *seg_details_y) {
 //
 // Allocates and loads all the structures needed for fast lookups of the   
 // index of an rr_node.  
 // rr_node_indices is a matrix containing the index  
 // of the *first* rr_node at a given (i,j) location.  
 // The chanx_rr_indices and chany_rr_indices data structures give 
 // the  rr_node  index for each track in each (i,j,) channel segment.                          
  
 int index, i, j;
 int **rr_node_indices;  
 
 rr_node_indices = (int **) alloc_matrix (0, nx+1, 0, ny+1, sizeof(int));
 
 chanx_rr_indices = (int ***) alloc_matrix3 (1, nx, 0, ny, 0, 
                    nodes_per_chan - 1, sizeof (int)); 
 
 chany_rr_indices = (int ***) alloc_matrix3 (0, nx, 1, ny, 0,
                    nodes_per_chan - 1, sizeof (int)); 
  // cristinel.ababei
 // chanx_rr_indices, chany_rr_indices: give  rr_node  index for 
 // each track in each (i,j) channel segment.
 // They will be heavily used inside get_rr_node_index().

 index = 0;
 
 for (i=0;i<=nx+1;i++) { 
    for (j=0;j<=ny+1;j++) {
       rr_node_indices[i][j] = index;
       // Deci asta e folosit ca un fel de referinta
       // la care se va add some sort of offset to find out
       // the index of some routing-segment, which starts
       // at location i,j and is on some of the tracks of
       // this channel.
       // oBs: rr_node_indices[i][j] vor contine numere de genul 0 7 15...
       // ptr ca index de fapt sweeps all rr_nodes created for all clbs 
       // (each clb is represented by nodes_per_clb) insa numai "inceputul"
       // unei secvente de rr_node (coresp to clb and its channels) este stored
       // in rr_node_indices[i][j];  then there are also created
       // rr_nodes for channels x and y at the location i,j.


       if (clb[i][j].type == CLB) { 
          index += nodes_per_clb; 
          index = load_chanx_rr_indices (seg_details_x, nodes_per_chan, index, 
                                         i, j);
          index = load_chany_rr_indices (seg_details_y, nodes_per_chan, index,
                                         i, j);
       } 

       else if (clb[i][j].type == IO) {  
          index += nodes_per_pad;

          if (j == 0)    /* Bottom row */
             index = load_chanx_rr_indices (seg_details_x, nodes_per_chan, 
                        index, i, j);

          if (i == 0)    /* Leftmost column */
             index = load_chany_rr_indices (seg_details_y, nodes_per_chan, 
                        index, i, j);
       } 

       else if (clb[i][j].type != ILLEGAL) {
          printf("Error in alloc_and_load_rr_node_indices.  Unexpected clb"
                 " type.\n");
          exit (1);
       } 
    }  
 }     
  printf(" | rr_nodes | = %d.\n",index);
 return (rr_node_indices);
}//eof
// ======================================================================= //



void free_rr_node_indices (int **rr_node_indices) {

/* Frees all the rr_node_indices structures allocated for fast index       *
 * computations.                                                           */

 free_matrix (rr_node_indices, 0, nx+1, 0, sizeof(int));
 free_matrix3 (chanx_rr_indices, 1, nx, 0, ny, 0, sizeof (int));
 free_matrix3 (chany_rr_indices, 0, nx, 1, ny, 0, sizeof (int));
}
// ======================================================================= //


static int load_chanx_rr_indices (t_seg_details *seg_details_x, int 
           nodes_per_chan, int start_index, int i, int j) {

/* Loads the chanx_rr_indices array for all track segments starting at     *
 * (i,j), assuming the first segment found has index start_index.  It also *
 * returns the next free index (i.e. the index for the next rr_node).      */

 int rr_index, istart, iend, iseg, itrack;

 rr_index = start_index;

 for (itrack=0;itrack<nodes_per_chan;itrack++) {
    istart = get_closest_seg_start (seg_details_x, itrack, i, j);

    /* Don't do anything if this isn't the start of the segment. */
 
    if (istart != i)
       continue;
   
    iend = get_seg_end (seg_details_x, itrack, istart, j, nx);
   
    for (iseg=istart;iseg<=iend;iseg++) 
       chanx_rr_indices[iseg][j][itrack] = rr_index;

    rr_index++;
 }
 
 return (rr_index);
}
// ======================================================================= //


static int load_chany_rr_indices (t_seg_details *seg_details_y, int 
           nodes_per_chan, int start_index, int i, int j) {
 
/* Loads the chany_rr_indices array for all track segments starting at     *
 * (i,j), assuming the first segment found has index start_index.  It also *
 * returns the next free index (i.e. the index for the next rr_node).      */
 
 int rr_index, jstart, jend, jseg, itrack;

 rr_index = start_index;
 
 for (itrack=0;itrack<nodes_per_chan;itrack++) {
    jstart = get_closest_seg_start (seg_details_y, itrack, j, i);
 
    /* Don't do anything if this isn't the start of the segment. */
  
    if (jstart != j)
       continue; 
    
    jend = get_seg_end (seg_details_y, itrack, jstart, i, ny);
    
    for (jseg=jstart;jseg<=jend;jseg++) 
       chany_rr_indices[i][jseg][itrack] = rr_index;  
 
    rr_index++; 
 } 
  
 return (rr_index); 
} 
// ======================================================================= //


int get_rr_node_index (int i, int j, t_rr_type rr_type, int ioff,
        int nodes_per_chan, int **rr_node_indices) {

/* Returns the index of the specified routing resource node.  (i,j) are     *
 * the location within the FPGA, rr_type specifies the type of resource,    *
 * and ioff gives the number of this resource.  ioff is the class number,   *
 * pin number or track number, depending on what type of resource this      *
 * is.  All ioffs start at 0 and go up to pins_per_clb-1 or the equivalent. *
 * The order within a clb is:  SOURCEs + SINKs (num_class of them); IPINs,  *
 * and OPINs (pins_per_clb of them); CHANX; and CHANY (nodes_per_chan of    *
 * each).  For (i,j) locations that point at pads the order is:  io_rat     *
 * occurances of SOURCE, SINK, OPIN, IPIN (one for each pad), then one      *
 * associated channel (if there is a channel at (i,j)).  All IO pads are    *
 * bidirectional, so while each will be used only as an INPAD or as an      *
 * OUTPAD, all the switches necessary to do both must be in each pad.       *
 *                                                                          *
 * Note that for segments (CHANX and CHANY) of length > 1, the segment is   *
 * given an rr_index based on the (i,j) location at which it starts (i.e.   *
 * lowest (i,j) location at which this segment exists).                     *
 * This routine also performs error checking to make sure the node in       *
 * question exists.                                                         */

 int index, iclass;

 assert (ioff >= 0);
 assert (i >= 0 && i < nx + 2);
 assert (j >= 0 && j < ny + 2);

 index = rr_node_indices[i][j];  /* Start of that block */
 
 switch (clb[i][j].type) {
 
 case CLB:
    switch (rr_type) {
 
    case SOURCE:
       assert (ioff < num_class);
       assert (class_inf[ioff].type == DRIVER);
 
       index += ioff;
       return (index);
  
    case SINK:
       assert (ioff < num_class);
       assert (class_inf[ioff].type == RECEIVER);
  
       index += ioff;
       return (index);
 
    case OPIN:
       assert (ioff < pins_per_clb);
       iclass = clb_pin_class[ioff];
       assert (class_inf[iclass].type == DRIVER);
 
       index += num_class + ioff;
       return (index);
 
    case IPIN:
       assert (ioff < pins_per_clb);
       iclass = clb_pin_class[ioff];
       assert (class_inf[iclass].type == RECEIVER);
 
       index += num_class + ioff;
       return (index);
 
    case CHANX:
       assert (ioff < nodes_per_chan);
       index = chanx_rr_indices[i][j][ioff];
       return (index);
 
    case CHANY:
       assert (ioff < nodes_per_chan);
       index = chany_rr_indices[i][j][ioff];
       return (index);
 
    default:
       printf ("Error:  Bad rr_node passed to get_rr_node_index.\n"
               "Request for type %d number %d at (%d, %d).\n", rr_type,
               ioff, i, j);
       exit (1);
    }
    break;
 
 case IO:
    switch (rr_type) {
 
    case SOURCE:
       assert (ioff < io_rat);
       index += 4 * ioff;
       return (index);
 
    case SINK:
       assert (ioff < io_rat);
       index += 4 * ioff + 1;
       return (index);
 
    case OPIN:
       assert (ioff < io_rat);
       index += 4 * ioff + 2;
       return (index);
 
    case IPIN:
       assert (ioff < io_rat);
       index += 4 * ioff + 3;
       return (index);
 
    case CHANX:
       assert (ioff < nodes_per_chan);
       assert (j == 0);                 /* Only one with a channel. */
       index = chanx_rr_indices[i][j][ioff];
       return (index);
 
    case CHANY:
       assert (ioff < nodes_per_chan);
       assert (i == 0);                 /* Only one with a channel. */
       index = chany_rr_indices[i][j][ioff];
       return (index);
 
    default:
       printf ("Error:  Bad rr_node passed to get_rr_node_index.\n"
               "Request for type %d number %d at (%d, %d).\n", rr_type,
               ioff, i, j);
       exit (1);
    }
    break;
 
 default:
    printf("Error in get_rr_node_index:  unexpected block type (%d) at "
           "(%d, %d).\nrr_type: %d.\n", clb[i][j].type, i, j, rr_type);
    exit (1);
 }
}
// ======================================================================= //


int get_seg_end (t_seg_details *seg_details, int itrack, int seg_start, int
       chan_num, int max_end) 
{
/* Returns the segment number (coordinate along the channel) at which this *
 * segment ends.  For a segment spanning clbs from 1 to 4, seg_end is 4.   *
 * max_end is the maximum dimension of the FPGA in this direction --       *
 * either nx or ny.                                                        */

 int seg_end, length, norm_start;

 length = seg_details[itrack].length;

 if (seg_start != 1) {
    seg_end = min (seg_start + length - 1, max_end);
 }
 else if (seg_details[itrack].longline) {
    seg_end = max_end;
 }
 else {        /* Short segment, at the left edge of array. */
    norm_start = seg_details[itrack].start;
    seg_end = length - (length + 1 + chan_num - norm_start) % length;
 }
 
 return (seg_end);
}//eof
// ======================================================================= //


int get_xtrack_to_clb_ipin_edges (int tr_istart, int tr_iend, int tr_j,
       int itrack, int iside, t_linked_edge **edge_list_ptr, struct s_ivec 
       **tracks_to_clb_ipin, int nodes_per_chan, int **rr_node_indices, 
       t_seg_details *seg_details_x, int wire_to_ipin_switch) {

/* This routine counts how many connections should be made from this segment *
 * to the clbs above (TOP) or below (BOTTOM) it.   It also adds them to the  *
 * edge linked list and updates edge_list_ptr.                               */

 int clb_j, iconn, num_conn, max_conn, i, ipin, to_node;
 t_linked_edge *edge_list_head;

/* Side is from the *track's* perspective */

 if (iside == BOTTOM) {
    clb_j = tr_j;
 }
 else if (iside == TOP) {
    clb_j = tr_j + 1;
 }
 else {  
    printf ("Error in get_xtrack_to_clb_ipin_edges:  Unknown iside: %d.\n",
            iside);
    exit (1);
 }

 edge_list_head = *edge_list_ptr;
 num_conn = 0;
 max_conn = tracks_to_clb_ipin[itrack][iside].nelem;

 for (i=tr_istart;i<=tr_iend;i++) {
    if (is_cbox (i, tr_j, itrack, seg_details_x)) {
       for (iconn=0;iconn<max_conn;iconn++) {
          ipin = tracks_to_clb_ipin[itrack][iside].list[iconn];
          to_node = get_rr_node_index (i, clb_j, IPIN, ipin, nodes_per_chan,
                    rr_node_indices);
          edge_list_head = insert_in_edge_list (edge_list_head, to_node,
                    wire_to_ipin_switch, &free_edge_list_head);
       }
       num_conn += max_conn;
    }
 }

 *edge_list_ptr = edge_list_head;
 return (num_conn);
}
// ======================================================================= //



int get_ytrack_to_clb_ipin_edges (int tr_jstart, int tr_jend, int tr_i,
       int itrack, int iside, t_linked_edge **edge_list_ptr, struct s_ivec
       **tracks_to_clb_ipin, int nodes_per_chan, int **rr_node_indices,
       t_seg_details *seg_details_y, int wire_to_ipin_switch) {

/* This routine counts how many connections should be made from this segment *
 * to the clbs to the LEFT or RIGHT of it.  It also adds them to the edge    *
 * linked list and updates edge_list_ptr.                                    */

 int clb_i, iconn, num_conn, max_conn, j, ipin, to_node;
 t_linked_edge *edge_list_head;

/* Side is from the *track's* perspective */

 if (iside == LEFT) {  
    clb_i = tr_i;
 }
 else if (iside == RIGHT) {
    clb_i = tr_i + 1;   
 }
 else {  
    printf ("Error in get_ytrack_to_clb_ipin_edges:  Unknown iside: %d.\n",
            iside);
    exit (1);
 }
 
 edge_list_head = *edge_list_ptr;
 num_conn = 0;   
 max_conn = tracks_to_clb_ipin[itrack][iside].nelem;
 
 for (j=tr_jstart;j<=tr_jend;j++) {
    if (is_cbox (j, tr_i, itrack, seg_details_y)) {
       for (iconn=0;iconn<max_conn;iconn++) {
          ipin = tracks_to_clb_ipin[itrack][iside].list[iconn];
          to_node = get_rr_node_index (clb_i, j, IPIN, ipin, nodes_per_chan,
                    rr_node_indices);
          edge_list_head = insert_in_edge_list (edge_list_head, to_node, 
                        wire_to_ipin_switch, &free_edge_list_head);
       }
       num_conn += max_conn;
    }   
 }
 
 *edge_list_ptr = edge_list_head;
 return (num_conn); 
}
// ======================================================================= //



int get_xtrack_to_pad_edges (int tr_istart, int tr_iend, int tr_j, int pad_j,
        int itrack, t_linked_edge **edge_list_ptr, struct s_ivec 
        *tracks_to_pads, int nodes_per_chan, int **rr_node_indices, 
        t_seg_details *seg_details_x, int wire_to_ipin_switch) {

/* This routine counts how many connections should be made from this segment *
 * to the row of pads above or below it.  It also adds these edges to the    *
 * edge list and updates edge_list_ptr to point to the new list head.        */

 int iconn, ipad, to_node, num_conn, max_conn, i;
 t_linked_edge *edge_list_head;

 edge_list_head = *edge_list_ptr;
 num_conn = 0;
 max_conn = tracks_to_pads[itrack].nelem;

 for (i=tr_istart;i<=tr_iend;i++) {
    if (is_cbox (i, tr_j, itrack, seg_details_x)) {
       for (iconn=0;iconn<max_conn;iconn++) {
          ipad = tracks_to_pads[itrack].list[iconn];
          to_node = get_rr_node_index (i, pad_j, IPIN, ipad, nodes_per_chan, 
                        rr_node_indices);
          edge_list_head = insert_in_edge_list (edge_list_head, to_node,
                           wire_to_ipin_switch, &free_edge_list_head);
       }
       num_conn += max_conn; 
    }
 }
 
 *edge_list_ptr = edge_list_head;
 return (num_conn);
}
// ======================================================================= //



int get_ytrack_to_pad_edges (int tr_jstart, int tr_jend, int tr_i, int pad_i, 
        int itrack, t_linked_edge **edge_list_ptr, struct s_ivec 
        *tracks_to_pads, int nodes_per_chan, int **rr_node_indices, 
        t_seg_details *seg_details_y, int wire_to_ipin_switch) {

/* This routine counts how many connections should be made from this segment *
 * to the row of pads to the left or right of it.  Additionally, if rr_edges *
 * isn't NULL, it will also load the edges array.  Make sure rr_edges points *
 * at the next free spot to put an edge in this case.                        */

 int iconn, ipad, to_node, num_conn, max_conn, j;
 t_linked_edge *edge_list_head;

 edge_list_head = *edge_list_ptr;
 num_conn = 0;
 max_conn = tracks_to_pads[itrack].nelem;

 for (j=tr_jstart;j<=tr_jend;j++) {
    if (is_cbox (j, tr_i, itrack, seg_details_y)) {
       for (iconn=0;iconn<tracks_to_pads[itrack].nelem;iconn++) {
          ipad = tracks_to_pads[itrack].list[iconn];
          to_node = get_rr_node_index (pad_i, j, IPIN, ipad, nodes_per_chan, 
                        rr_node_indices);
          edge_list_head = insert_in_edge_list (edge_list_head, to_node,
                           wire_to_ipin_switch, &free_edge_list_head);
       } 
       num_conn += max_conn;
    }  
 }
 
 *edge_list_ptr = edge_list_head; 
 return (num_conn);
}

// ======================================================================= //
int get_xtrack_to_ytracks (int from_istart, int from_iend, int from_j, int 
       from_track, int to_j, t_linked_edge **edge_list_ptr, int nodes_per_chan,
       int **rr_node_indices, t_seg_details *seg_details_x, t_seg_details 
       *seg_details_y, enum e_switch_block_type switch_block_type,
			   boolean during_lut_generation) 
{
// cristinel.ababei
// 
// oBs: from_j == to_j
//
/* Counts how many connections should be made from this segment to the y-   *
 * segments in the adjacent channels at to_j.  It returns the number of     *
 * connections, and updates edge_list_ptr to point at the head of the       *
 * (extended) linked list giving the nodes to which this segment connects   *
 * and the switch type used to connect to each.                             *
 *                                                                          *
 * An edge is added from this segment to a y-segment if:                    *
 * (1) this segment should have a switch box at that location, or           *
 * (2) the y-segment to which it would connect has a switch box, and the    *
 *     switch type of that y-segment is unbuffered (bidirectional pass      *
 *     transistor).                                                         *
 *                                                                          *
 * If the switch in each direction is a pass transistor (unbuffered), both  *
 * switches are marked as being of the types of the larger (lower R) pass   *
 * transistor.  Note that this code implicitly assumes the routing is       *
 * bidirectional (a switch in each direction).                              */

// Here, the process of finding all connections which involve routing segment
// which starts at (from_istart,from_j) takes place.  I have to count 
// all connections from groups  Group123456 and  Group7 in same manner as
// they were counted with "virtual" fucntions derived from these (those
// used to finally compute insertion-steps)

 int num_conn, to_node, i, to_track, iconn;
 int from_node_switch, to_node_switch;
 short switch_types[2];
 boolean is_x_sbox, is_y_sbox, yconn_to_above;
 t_linked_edge *edge_list_head;
 struct s_ivec conn_tracks;
 enum e_hwrp_switch_types check_connection_type=NORMAL_SWITCH;//hwrp's.
 //boolean should_be_skipped=TRUE;//hwrp's.

 num_conn = 0;
 edge_list_head = *edge_list_ptr;// List where edges (1 or 2) are added.

 if (to_j <= from_j) {
   yconn_to_above = TRUE;  /* The connection goes up from ychan to xchan. */
 }
 else {
   yconn_to_above = FALSE;
 }


 // Recall:  this is the type of switch to use on switches that go *to* this
 // node (i.e. the output side is on the from_node).    
 from_node_switch = seg_details_x[from_track].wire_switch;
 // Asta .wire_switch aici should tell me that "to" this track connections
 // (if any) should be made using hard wired switches.  Based on from_track
 // seg_details_x[] will tell me that its a T or L or etc. hwrp based on 
 // a simple strategy de a distribui hwrp's accros the FPGA area.  SB's
 // are not identical (because of mixed wire segment lengths and staggering)
 // and therefore i might have a L hwrp at location (i,j) but not at (i+1,j)
 // etc.

/* For each unit-length piece of wire in the segment, I add the things it    *
 * would connect to diagonally to the right and diagonally to the left of    *
 * it.  This is important -- for some switch boxes, the connection from      *
 * (i,j) to (i,j+1) and the connection from (i+1,j) to (i,j+1), for example, *
 * would lead to connections to different tracks in the (i,j+1) channel.  In *
 * this case, the code below will connect to ALL the tracks that two unit    *
 * length segments at (i,j) and (i+1,j) would have connected to in (i,j+1).  */

 for (i=from_istart;i<=from_iend;i++) { 
   // Acum mergem de-a lungul wire-segmentului de la "from_start"
   // till "from_end" si for every location ma uit to see whether 
   // there should be a switch "on" this segment and if so I have to "look" at 
   // possible connections to tracks (wire-segments) in y channels.

   // ---------       ---------
   // SB      |_seg___|      SB
   //        /|       |\             And we go from smaller i to larger i
   //       / |       | \            sweeping the roting segment from "from_istart"
   //                                till "from_iend"
   //  LEFT      :-)      RIGHT
   // (i-1,j)   (i,j)     (i,j)
   //
   // 
   //        __________|____SB
   //       |        __|__  |
   //    ___|__| T1    T2   |
   //       |  |          |_|___T3
   //       |             | |
   //       |  |      L2|   |
   //    ___|__| L1     |___|___
   //       |        _______|___  
   //    ___|__     |  ___  |  L3
   //       |__|____|___|___|
   //          |L4  |   |T4


   // ------------------------------------------------------------------------
   // ===>  Diagonal connection to right (from xchan to ychan) 
   // -- TO RIGHT
   //
   // oBs:  This connection "4" has to be processed first among all (1,2,3,4)
   // in order to go the same way as when I counted all conections with "virtual" 
   // functions.  Thus, once a connection is detected all others (1,2,3) will not
   // be let go as usual (let things be constructed as in normal VPR) but they
   // will be processed now (at detection time) and made to implement the type
   // of hard-wired-routing-pattern as required by their frequency.
   //   ___
   // 1|/\2|
   //  |\ \| 
   //  |4\/|3
   //   --- 
   is_x_sbox = is_sbox (i, from_j, from_track, seg_details_x, TRUE);
   // Adica connexiuni in jos, "in fata" as we go de-a lungul 
   // segmentului.
   conn_tracks = get_switch_box_tracks (
		i, from_j,  // Loc. on CHANX during mersului.
		from_track, // Inside xchan
		CHANX, 
		i, to_j,    // Which y-chan location
		CHANY, 
		switch_block_type, nodes_per_chan);
  
   /* For all the tracks we connect to in that channel ... */
   // Normally, for simple switch-boxes is only one...
   for (iconn=0;iconn<conn_tracks.nelem;iconn++) {
     to_track = conn_tracks.list[iconn];
     is_y_sbox = is_sbox (to_j, i, to_track, seg_details_y, yconn_to_above);
     to_node_switch = seg_details_y[to_track].wire_switch;
       
     get_switch_type (is_x_sbox, is_y_sbox, from_node_switch, to_node_switch,
		      switch_types);

     // ...................................................................
     if (switch_types[0] != OPEN) { // Roughly this only tells me that 
                                    // there is a connection here.

       to_node = get_rr_node_index (i, to_j, CHANY, to_track, 
				    nodes_per_chan, rr_node_indices);


       if (!rr_edge_done[to_node]) {        /* Not a repeat edge. */

	 // --- c.a:  Now, I am at the moment when I have to see if this detected 
	 // connection either as  Group123456 and  Group7 has to be made
	 // as hard-wired-routing-pattern inside the SB and what type 
	 // H, V, CROSS, L or T
	 // As I look to a potential SB to the "right" along this segment 
	 // on this track.
	 if ( to_j <= from_j ) { // yconn_to_above = TRUE
	   // This is the "case" in the way I did counting.
	   check_connection_type = get_switch_type_HWRP(i, to_j, 
				   from_track, nodes_per_chan);
	 }
	 else {                  // yconn_to_above = FALSE
	   check_connection_type = get_switch_type_HWRP(i, (to_j-1), 
				   from_track, nodes_per_chan);
	 }
	 // The above will make sure that below I will or will not continue
	 // inserting normal conn-switches as in normal VPR for
	 // this track at this location, because once a connection in
	 // Group1-6 or Group7 was detected all 4 or 2 terminals
	 // participate in hwrp.  Though this could be relaxed.
	 // Next, at this time we are during delay lut's creation (done
	 // only once at the begining of VPR run) then skip the whole stuff
	 // related to hwrp and do it as in normal VPR:
	 if ( during_lut_generation ) check_connection_type = NORMAL_SWITCH;
	 // ......................................................................
	 //
	 if ( check_connection_type == NORMAL_SWITCH ) {
	   // Do not do anything special.  Just let things "go"
	   // as in normal VPR:  build rr_graph edges as usual
	   // one by one as we sweep all channels.
	   num_conn++;
	   rr_edge_done[to_node] = TRUE;
	   edge_list_head = insert_in_edge_list (edge_list_head, to_node,
				   switch_types[0], &free_edge_list_head);
	   if (switch_types[1] != OPEN) {    /* A second edge */
	     edge_list_head = insert_in_edge_list (edge_list_head, to_node,
				     switch_types[1], &free_edge_list_head);
	     num_conn++;
	   }
	 }
	 // ......................................................................
	 // Case hwrp's:  HORIZONTAL, VERTICAL, CROSS, L, T, etc.
	 else {
	   // Well, this is special case when we have to do a little
	   // more.  Here I see the required type of hwrp (as returned
	   // by get_switch_type_HWRP above) and build the entire
	   // connection at this detection time.
	   //printf("a");
	   if ( to_j <= from_j ) { // yconn_to_above = TRUE
	     if((check_connection_type == L4)||
		(check_connection_type == T1)|| 
		(check_connection_type == T4)) {
	       switch_types[0] = 1;
	       edge_list_head = insert_in_edge_list (edge_list_head, to_node,
				   switch_types[0], &free_edge_list_head);
	       num_conn++;
	     }
	   }
	   else {                  // yconn_to_above = FALSE
	     if((check_connection_type == L1)||
		(check_connection_type == T1)|| 
		(check_connection_type == T2)) {
	       switch_types[0] = 1;
	       edge_list_head = insert_in_edge_list (edge_list_head, to_node,
				   switch_types[0], &free_edge_list_head);
	       num_conn++;
	     }  
	   }

	 }//else
	 //
	 // ......................................................................

       }//if "Not a repeat edge."
     }//if
     // ...................................................................
   }//for
   // ------------------------------------------------------------------------

   // ------------------------------------------------------------------------
   // ===> Diagonal connection to left (from xchan to ychan)
   // -- TO LEFT
   // 
   // Next returns TRUE if the specified segment HAS a switch box at specified  
   // location, FALSE otherwise.  The switch box is from the specified segment 
   // to the left (for chanx) or below (for chany) unless above_or_right is    
   // true.
   is_x_sbox = is_sbox (i, from_j, from_track, seg_details_x, FALSE);
   // oBs:  is_x_sbox and is_y_sbox tell me if there is connection
   // at this location between chanx and chany, if it's chanx-->chany
   // or chanx<--chany or chanx<-->chany or not at all.  Type of 
   // connection is given by switch_type and any reconciliation will
   // be done inside get_switch_type() - which will extract
   // needed information from switch_inf[].
   // Next function returns a vector of the tracks to which from_track 
   // at (from_i,from_j) x-chan should connect at (i-1,to_j) y-chan.
   // Adica connexiuni in jos, "in spatele" as we go de-a lungul 
   // segmentului.
   conn_tracks = // Deci astea sint conexiuni in interiorul aceluiasi s-b.
     get_switch_box_tracks (
		i, from_j,  // Loc. on CHANX during mersului.
		from_track, // Inside xchan
		CHANX, 
		i-1, to_j,  // Which y-chan location
		CHANY, 
		switch_block_type, nodes_per_chan);
       

   /* For all the tracks we connect to in that channel ... */
   // For the simplest/standard switch-box a track connects only to its 
   // counterpart track because Fs=3. 
   // Hence next "for" loop will be gone thru once.

   for (iconn=0;iconn<conn_tracks.nelem;iconn++) {
     to_track = conn_tracks.list[iconn];
     // [  ?  ] Is this like a double-check; because I know already that
     // the track from x-chan is connected to a track in y-chan - that
     // is why "iconn" was added in the conn_tracks list, right? 
     // I see, in one direction I might connect with a type of switch
     // while in the other direction with a different type...
     is_y_sbox = is_sbox (to_j, i-1, to_track, seg_details_y, yconn_to_above);
     to_node_switch = seg_details_y[to_track].wire_switch;
 
     // This routine looks at whether the from_node and to_node want a switch,  
     // and what type of switch is used to connect *to* each type of node       
     // (from_node_switch and to_node_switch).  It decides what type of switch, 
     // if any, should be used to go from from_node to to_node.  If no switch   
     // should be inserted (i.e. no connection), it returns OPEN.  Its returned 
     // values are in the switch_types array.  It needs to return an array      
     // because one topology (a buffer in the forward direction and a pass      
     // transistor in the backward direction) results in *two* switches. 
     // Next will populate switch_types[0] and [1].  This will say
     // connection or not in one direction, the other or both.
     // So, from_node_switch,to_node_switch  tell me - for the moment -    
     // the type of switch (hwrp_switch, buffered or pass-trans)
     // if any, whereas is_x_sbox,is_y_sbox say if
     // there exists a Switch-Block inside which these connection(s)
     // should be made.  So in next function we will see if there
     // is aggreement and correct any conflicts and finally see
     // what type of switch to use. 
     get_switch_type (is_x_sbox, is_y_sbox, 
		      from_node_switch, to_node_switch,    
		      switch_types); // oBs:  Can be hwrp_switch type, or
                                     // buffered or pass-transistir.

     if (switch_types[0] != OPEN) {
       to_node = get_rr_node_index (i-1, to_j, CHANY, to_track, 
			   nodes_per_chan, rr_node_indices);

       if (!rr_edge_done[to_node]) {       /* Not a repeat edge. */


	 if ( to_j <= from_j ) { // yconn_to_above = TRUE
	   // This is the "case" in the way I did counting.
	   check_connection_type = get_switch_type_HWRP(i-1, to_j, 
				   (from_track + 2*nodes_per_chan), nodes_per_chan);
	 }
	 else {                  // yconn_to_above = FALSE
	   check_connection_type = get_switch_type_HWRP(i-1, (to_j-1), 
				   (from_track + 2*nodes_per_chan), nodes_per_chan);
	 }
	 // The above will make sure that below I will not continue
	 // inserting normal conn-switches as in normal VPR for
	 // this track at this location, because once a connection in
	 // Group1-6 or Group7 was detected all 4 or 2 terminals
	 // participate in hwrp.  Though this could be relaxed.
	 // Next, at this time we are during delay lut's creation (done
	 // only once at the begining of VPR run) then skip the whole stuff
	 // related to hwrp and do it as in normal VPR:
	 if ( during_lut_generation ) check_connection_type = NORMAL_SWITCH;
	 // ......................................................................
	 //
	 if ( check_connection_type == NORMAL_SWITCH ) {
	   // Do not do anything special.  Just let things "go"
	   // as in normal VPR:  build rr_graph edges as usual
	   // one by one as we sweep all channels
	   num_conn++;
	   rr_edge_done[to_node] = TRUE;
	   edge_list_head = insert_in_edge_list (edge_list_head, to_node,
			           switch_types[0], &free_edge_list_head);
	   if (switch_types[1] != OPEN) {   /* A second edge. */
	     edge_list_head = insert_in_edge_list (edge_list_head, to_node,
                                     switch_types[1], &free_edge_list_head);
	     num_conn++;
	   }
	 }
	 // ......................................................................
	 // Case hwrp's:  HORIZONTAL, VERTICAL, CROSS, L, T, etc.
	 else {
	   //printf("b");
	   if ( to_j <= from_j ) { // yconn_to_above = TRUE
	     if((check_connection_type == L3)||
		(check_connection_type == T3)|| 
		(check_connection_type == T4)) {
	       switch_types[0] = 1;
	       edge_list_head = insert_in_edge_list (edge_list_head, to_node,
				   switch_types[0], &free_edge_list_head);
	       num_conn++;
	     }
	   }
	   else {                  // yconn_to_above = FALSE
	     if((check_connection_type == L2)||
		(check_connection_type == T2)|| 
		(check_connection_type == T3)) {
	       switch_types[0] = 1;
	       edge_list_head = insert_in_edge_list (edge_list_head, to_node,
				   switch_types[0], &free_edge_list_head);
	       num_conn++;
	     }  
	   }

	 }//else
	 //
	 // ......................................................................


       }//if "Not a repeat edge."
       // Deci in momentul asta am adaugat cu acest if in edge_list_head 
       // 0 or 1 or 2 edges (bidirectional case) care imediat after the 
       // return from this routine, will get translated into edges of
       // rr_graph.  These edges will in this way model connections
       // made inside SB as normal connections (buffered or pass-trans)
       // or hard-wired patterns.
     }//if
   }//for
   // ------------------------------------------------------------------------




 }// for (i=from_istart;i<=from_iend;i++) ---> End for length of segment.
 
 *edge_list_ptr = edge_list_head;
 return (num_conn);
 // At this time we have added some rr_node idices (coresp. to routing-segments)
 // to the fanout list with connections of the wire-segment 
 // characterized by: begins at (from_istart,from_j) is located
 // on track "from_track" inside the channel x and extends till "to_j".
 // Note that this "mers de-a lungul lui x-chan" is in same spirit
 // as segment (i.e., rr_indices) indices were "counted".
}//eof
// ======================================================================= //



int get_ytrack_to_xtracks (int from_jstart, int from_jend, int from_i, int 
       from_track, int to_i, t_linked_edge **edge_list_ptr, int nodes_per_chan, 
       int **rr_node_indices, t_seg_details *seg_details_x, t_seg_details 
       *seg_details_y, enum e_switch_block_type switch_block_type,
			   boolean during_lut_generation) 
{
/* Counts how many connections should be made from this segment to the x-   *
 * segments in the adjacent channels at to_i.  It returns the number of     *
 * connections, and updates edge_list_ptr to point at the head of the       *
 * (extended) linked list giving the nodes to which this segment connects   *
 * and the switch type used to connect to each.                             *
 *                                                                          *
 * An edge is added from this segment to an x-segment if:                   *
 * (1) this segment should have a switch box at that location, or           *
 * (2) the x-segment to which it would connect has a switch box, and the    *
 *     switch type of that x-segment is unbuffered (bidirectional pass      *
 *     transistor).                                                         *
 *                                                                          *
 * If the switch in each direction is a pass transistor (unbuffered), both  *
 * switches are marked as being of the types of the larger (lower R) pass   *
 * transistor.  Note that this code implicitly assumes the routing is       *
 * bidirectional (a switch in each direction).                              */

 
 int num_conn, to_node, j, to_track, iconn;
 int from_node_switch, to_node_switch;
 short switch_types[2];
 boolean is_x_sbox, is_y_sbox, xconn_to_right;
 t_linked_edge *edge_list_head;
 struct s_ivec conn_tracks;
 enum e_hwrp_switch_types check_connection_type=NORMAL_SWITCH;//hwrp's.
 //boolean should_be_skipped=TRUE;//hwrp's.

 num_conn = 0;
 edge_list_head = *edge_list_ptr;

 if (to_i <= from_i) {
   xconn_to_right = TRUE;  /* The connection goes RIGHT from xchan to ychan */
 }
 else {
   xconn_to_right = FALSE;
 }

/* Recall:  this is the type of switch to use on switches that go *to* this *
 * node (i.e. the output side is on the from_node).                         */
 from_node_switch = seg_details_y[from_track].wire_switch;

/* For each unit-length piece of wire in the segment, I add the things it    *
 * would connect to diagonally above and diagonally below it.  This is       *
 * important -- for some switch boxes, the connection from (i,j) to (i+1,j)  *
 * and the connection from (i,j+1) to (i+1,j), for example, would lead to    *
 * connections to different tracks in the (i+1,j) channel.  In this case,    *
 * the code below will connect to ALL the tracks that two unit length        *
 * segments at (i,j) and (i,j+1) would have connected to in (i+1,j).         */


 for (j=from_jstart;j<=from_jend;j++) {

   // ------------------------------------------------------------------------
   // ===>  Diagonal connection to above (from ychan to xchan)
   // -- TO ABOVE
   //
   is_y_sbox = is_sbox (j, from_i, from_track, seg_details_y, TRUE);

   conn_tracks = get_switch_box_tracks (from_i, j, from_track, CHANY, to_i,
               j, CHANX, switch_block_type, nodes_per_chan);

   /* For all the tracks we connect to in that channel ... */ 

   for (iconn=0;iconn<conn_tracks.nelem;iconn++) { 
     to_track = conn_tracks.list[iconn];
     is_x_sbox = is_sbox (to_i, j, to_track, seg_details_x, xconn_to_right);
     to_node_switch = seg_details_x[to_track].wire_switch;

     get_switch_type (is_y_sbox, is_x_sbox, from_node_switch, to_node_switch,
		      switch_types);

     if (switch_types[0] != OPEN) { 
       to_node = get_rr_node_index (to_i, j, CHANX, to_track, 
				    nodes_per_chan, rr_node_indices);
       
       if (!rr_edge_done[to_node]) {       /* Not a repeat edge. */


	 if ( to_i <= from_i ) { // xconn_to_right = TRUE
	   check_connection_type = get_switch_type_HWRP((from_i-0), j, 
				   (from_track + 3*nodes_per_chan),nodes_per_chan);
	 }
	 else {                  // xconn_to_right = FALSE
	   check_connection_type = get_switch_type_HWRP((from_i+0), j, 
				   (from_track + 3*nodes_per_chan),nodes_per_chan);
	 }
	 // The above will make sure that below I will or will not continue
	 // inserting normal conn-switches as in normal VPR for
	 // this track at this location, because once a connection in
	 // Group1-6 or Group7 was detected, all 4 or 2 terminals
	 // participate in hwrp.  Though this could be relaxed.
	 // Next, at this time we are during delay lut's creation (done
	 // only once at the begining of VPR run) then skip the whole stuff
	 // related to hwrp and do it as in normal VPR:
	 if ( during_lut_generation ) check_connection_type = NORMAL_SWITCH;
	 // ......................................................................
	 //
	 if ( check_connection_type == NORMAL_SWITCH ) {
	   // Just let things "go" as in normal VPR
	   num_conn++; 
	   rr_edge_done[to_node] = TRUE;
	   edge_list_head = insert_in_edge_list (edge_list_head, to_node, 
				   switch_types[0], &free_edge_list_head);
	   if (switch_types[1] != OPEN) {  /* A second edge */
	     edge_list_head = insert_in_edge_list (edge_list_head, to_node, 
				     switch_types[1], &free_edge_list_head);
	     num_conn++;
	   }
	 }
	 // ......................................................................
	 // Case hwrp's:  HORIZONTAL, VERTICAL, CROSS, L, T, etc.
	 else {
	   //printf("c");
	   if ( to_i <= from_i ) { // xconn_to_right = TRUE
	     if((check_connection_type == L4)||
		(check_connection_type == T1)|| 
		(check_connection_type == T4)) {
	       switch_types[0] = 1;
	       edge_list_head = insert_in_edge_list (edge_list_head, to_node,
				   switch_types[0], &free_edge_list_head);
	       num_conn++;
	     }
	   }
	   else {                  // xconn_to_right = FALSE
	     if((check_connection_type == L3)||
		(check_connection_type == T3)|| 
		(check_connection_type == T4)) {
	       switch_types[0] = 1;
	       edge_list_head = insert_in_edge_list (edge_list_head, to_node,
				   switch_types[0], &free_edge_list_head);
	       num_conn++;
	     }  
	   }

	 }//else
	 //
	 // ......................................................................

       }//if "Not a repeat edge."

     }//if
   }//for
   // ------------------------------------------------------------------------

   // ------------------------------------------------------------------------
   // ===>  Diagonal connection to below (from ychan to xchan)
   // -- TO BELOW
   //
   is_y_sbox = is_sbox (j, from_i, from_track, seg_details_y, FALSE);

   conn_tracks = get_switch_box_tracks (from_i, j, from_track, CHANY, to_i,
               j-1, CHANX, switch_block_type, nodes_per_chan);

   /* For all the tracks we connect to in that channel ... */

   for (iconn=0;iconn<conn_tracks.nelem;iconn++) {
     to_track = conn_tracks.list[iconn];
     is_x_sbox = is_sbox (to_i, j-1, to_track, seg_details_x, xconn_to_right);
     to_node_switch = seg_details_x[to_track].wire_switch;
     
     get_switch_type (is_y_sbox, is_x_sbox, from_node_switch, to_node_switch,
                        switch_types);

     if (switch_types[0] != OPEN) { 
       to_node = get_rr_node_index (to_i, j-1, CHANX, to_track, 
                    nodes_per_chan, rr_node_indices);

       if (!rr_edge_done[to_node]) {       /* Not a repeat edge. */

	 if ( to_i <= from_i ) { // xconn_to_right = TRUE
	   check_connection_type = get_switch_type_HWRP((from_i-0), j-1, 
				   (from_track + 1*nodes_per_chan),nodes_per_chan);
	 }
	 else {                  // xconn_to_right = FALSE
	   check_connection_type = get_switch_type_HWRP((from_i-0), j-1, 
				   (from_track + 1*nodes_per_chan),nodes_per_chan);
	 }
	 // The above will make sure that below I will or will not continue
	 // inserting normal conn-switches as in normal VPR for
	 // this track at this location, because once a connection in
	 // Group1-6 or Group7 was detected, all 4 or 2 terminals
	 // participate in hwrp.  Though this could be relaxed.
	 // Next, at this time we are during delay lut's creation (done
	 // only once at the begining of VPR run) then skip the whole stuff
	 // related to hwrp and do it as in normal VPR:
	 if ( during_lut_generation ) check_connection_type = NORMAL_SWITCH;
	 // ......................................................................
	 //
	 if ( check_connection_type == NORMAL_SWITCH ) {
	   // Just let things "go" as in normal VPR
	   num_conn++; 
	   rr_edge_done[to_node] = TRUE;
	   edge_list_head = insert_in_edge_list (edge_list_head, to_node,
				   switch_types[0], &free_edge_list_head);
	   if (switch_types[1] != OPEN) {   /* A second edge. */
	     edge_list_head = insert_in_edge_list (edge_list_head, to_node,
				     switch_types[1], &free_edge_list_head);
	     num_conn++;
	   }
	 }
	 // ......................................................................
	 // Case hwrp's:  HORIZONTAL, VERTICAL, CROSS, L, T, etc.
	 else {
	   //printf("d");
	   if ( to_i <= from_i ) { // xconn_to_right = TRUE
	     if((check_connection_type == L1)||
		(check_connection_type == T1)|| 
		(check_connection_type == T2)) {
	       switch_types[0] = 1;
	       edge_list_head = insert_in_edge_list (edge_list_head, to_node,
				   switch_types[0], &free_edge_list_head);
	       num_conn++;
	     }
	   }
	   else {                  // xconn_to_right = FALSE
	     if((check_connection_type == L2)||
		(check_connection_type == T2)|| 
		(check_connection_type == T3)) {
	       switch_types[0] = 1;
	       edge_list_head = insert_in_edge_list (edge_list_head, to_node,
				   switch_types[0], &free_edge_list_head);
	       num_conn++;
	     }  
	   }

	 }//else
	 //
	 // ......................................................................

       }//if "Not a repeat edge."

     }//if
   }//for
   // ------------------------------------------------------------------------



 }// for (j=from_jstart;j<=from_jend;j++) ---> End for length of segment.
    
 
 *edge_list_ptr = edge_list_head;
 return (num_conn);
}//eof
// ======================================================================= //


int get_xtrack_to_xtrack (int from_i, int j, int from_track, int to_i, 
       t_linked_edge **edge_list_ptr, int nodes_per_chan, int 
       **rr_node_indices, t_seg_details *seg_details_x, enum 
       e_switch_block_type switch_block_type,
			  boolean during_lut_generation) 
{
// Returns the number of edges between the specified channel segments.      
// Also updates edge_list_ptr to point at the new (extended) linked list    
// of edges and switch types.
//
// cristinel.ababei
// Deci links between a track (wire-segment) within current
// channel and tracks in "continuation" adjacent channels. For x(y)
// channels these links are to the left(bottom)/right(right) x(y) channel.  
// These links with afferent edges being created, are made based on the 
// switch-box information.

 boolean is_from_sbox, is_to_sbox, from_goes_right, to_goes_right;
 int to_track, to_node, iconn, num_conn;
 int from_node_switch, to_node_switch;
 short switch_types[2];
 struct s_ivec conn_tracks;
 enum e_hwrp_switch_types check_connection_type=NORMAL_SWITCH;//hwrp's.

 if (from_i < to_i) {
   from_goes_right = TRUE;  // from_i == iend : right-most end coordinate. 
   to_goes_right = FALSE;   // to_i   == iend + 1 
 }
 else {
   from_goes_right = FALSE; // from_i == istart : left-most end coordinate. 
   to_goes_right = TRUE;    // to_i   == istart - 1 
 }

 num_conn = 0;
 from_node_switch = seg_details_x[from_track].wire_switch;

 is_from_sbox = is_sbox (from_i, j, from_track, seg_details_x, from_goes_right);
 // Next function returns a vector of the tracks to which from_track 
 // at (from_i, j) should connect at (to_i, j).
 //
 // cristinel.ababei:  What returns will also contain vertical tracks!?
 //
 conn_tracks = get_switch_box_tracks (from_i, j, from_track, CHANX, to_i,
            j, CHANX, switch_block_type, nodes_per_chan);

 /* For all the tracks we connect to in that channel ... */
 
 for (iconn=0;iconn<conn_tracks.nelem;iconn++) {
    to_track = conn_tracks.list[iconn];
    is_to_sbox = is_sbox (to_i, j, to_track, seg_details_x, to_goes_right);
    to_node_switch = seg_details_x[to_track].wire_switch;
  
    get_switch_type (is_from_sbox, is_to_sbox, from_node_switch,
                     to_node_switch, switch_types);
 
    if (switch_types[0] != OPEN) {
       to_node = get_rr_node_index (to_i, j, CHANX, to_track, 
				    nodes_per_chan, rr_node_indices);
       /* No need to check for repeats with the current switch boxes. */

       // c.a: Based on how  from_i<-->to_i I look and check the SB
       // for switch-connection type.
       if ( from_i < to_i ){ // from_goes_right = TRUE
	 check_connection_type = get_switch_type_HWRP(from_i, j, 
				 from_track, nodes_per_chan);
       }
       else {                // to_goes_right = TRUE
	 check_connection_type = get_switch_type_HWRP(to_i, j, 
				 (from_track + 2*nodes_per_chan), nodes_per_chan);
       }
       // The above will make sure that below I will or will not continue
       // inserting normal conn-switches as in normal VPR for
       // this track at this location, because once a connection in
       // Group1-6 or Group7 was detected, all 4 or 2 terminals
       // participate in hwrp.  Though this could be relaxed.
       // Next, at this time we are during delay lut's creation (done
       // only once at the begining of VPR run) then skip the whole stuff
       // related to hwrp and do it as in normal VPR:
       if ( during_lut_generation ) check_connection_type = NORMAL_SWITCH;
       // ......................................................................
       //
       if ( (check_connection_type == NORMAL_SWITCH)||
	    (check_connection_type == T1)|| 
	    (check_connection_type == T3) )
	 {
	   // Just let things "go" as in normal VPR if it's a normal switch
	   // OR T1, T3 because in this case the T1, T3 will be implemented
	   // as hard-wired but the fourth terminal will be connected
	   // "normally" otherwise would be dangling in the air and result 
	   // in rr_graph checking errors. So T's will be hard-wired T's
	   // with their fourt terminal connected normally.
	   // E.g.:
	   //  
	   //      __|__
	   // T1  |  |  |
	   //   -----|XXXX Fourth terminal of T1 connected normally.
	   //     |__|__|  in order not to remain in the air, unconnected
	   //        |     with no fanin or fanout.
	   //
	   *edge_list_ptr = insert_in_edge_list (*edge_list_ptr, to_node, 
                        switch_types[0], &free_edge_list_head);
	   num_conn++;
	   if (switch_types[1] != OPEN) {
	     *edge_list_ptr = insert_in_edge_list (*edge_list_ptr, to_node, 
                           switch_types[1], &free_edge_list_head);
	     num_conn++;
	   }
	 }
       // ......................................................................
       // Case hwrp's:  HORIZONTAL, VERTICAL, CROSS, L, T, etc.
       else {
	 //printf(" %d,%d ",from_i,j);
	 //printf("%d",check_connection_type);
	 if((check_connection_type == HORIZONTAL)||
	    (check_connection_type == T2)|| 
	    (check_connection_type == T4)) {
	   //printf("h");
	   // Create horizontal hwrp.
	   // oBs:  All hwrp are bi-directional because they play role 
	   // of electrically short-circuits.  This connection will be
	   // made same type in the opposite direction when I will 
	   // sweep to_node --> from_node.  Now, over-write switch_types[]
	   // with appropriate type.
	   switch_types[0] = 1; // switch_hwrp
	   *edge_list_ptr = insert_in_edge_list (*edge_list_ptr, to_node,
				   switch_types[0], &free_edge_list_head);
	   num_conn++;
	 }
	 else if ((check_connection_type == T1)|| 
		  (check_connection_type == T3)){

	 }
       }
       //
       // ......................................................................

    }//if
 }//for   
 
 return (num_conn);
}//eof


// ======================================================================= //


int get_ytrack_to_ytrack (int i, int from_j, int from_track, int to_j,
       t_linked_edge **edge_list_ptr, int nodes_per_chan, int 
       **rr_node_indices, t_seg_details *seg_details_y, enum 
       e_switch_block_type switch_block_type,
			  boolean during_lut_generation) 
{
// Returns the number of edges between the specified channel segments.     
// Also updates edge_list_ptr to point at the new (extended) linked list   
// of edges and switch types.          
 
 boolean is_from_sbox, is_to_sbox, from_goes_up, to_goes_up; 
 int to_track, to_node, iconn, num_conn;
 int from_node_switch, to_node_switch;
 short switch_types[2];
 struct s_ivec conn_tracks;
 enum e_hwrp_switch_types check_connection_type=NORMAL_SWITCH;//hwrp's.

 if (from_j < to_j) { 
   from_goes_up = TRUE;   // from_j == jend : top-most end coordinate. 
   to_goes_up = FALSE;    // to_j   == jend + 1 
 } 
 else {
   from_goes_up = FALSE;  // from_j == jstart : bottom-most end coordinate. 
   to_goes_up = TRUE;     // to_j   == jstart - 1 
 }


 num_conn = 0;
 from_node_switch = seg_details_y[from_track].wire_switch;
 
 is_from_sbox = is_sbox (from_j, i, from_track, seg_details_y, from_goes_up); 
 conn_tracks = get_switch_box_tracks (i, from_j, from_track, CHANY, i, to_j, 
             CHANY, switch_block_type, nodes_per_chan);
 
 /* For all the tracks we connect to in that channel ... */
 
 for (iconn=0;iconn<conn_tracks.nelem;iconn++) {
    to_track = conn_tracks.list[iconn];
    is_to_sbox = is_sbox (to_j, i, to_track, seg_details_y, to_goes_up); 
    to_node_switch = seg_details_y[to_track].wire_switch;
 
    get_switch_type (is_from_sbox, is_to_sbox, from_node_switch,
                     to_node_switch, switch_types);

    if (switch_types[0] != OPEN) { 
       to_node = get_rr_node_index (i, to_j, CHANY, to_track, 
				    nodes_per_chan, rr_node_indices); 

       /* No need to check for repeats with the current switch boxes. */
       // Based on how  from_i<-->to_i I look and check the SB
       // for switch-connection type.

       if ( from_j < to_j ){ // from_goes_up = TRUE
	 check_connection_type = get_switch_type_HWRP(i, from_j, 
				 (from_track + 3*nodes_per_chan), nodes_per_chan);
	 //from_track, nodes_per_chan);
       }
       else {                // to_goes_up = TRUE
	 check_connection_type = get_switch_type_HWRP(i, to_j, 
				 (from_track + 1*nodes_per_chan), nodes_per_chan);
	 //from_track, nodes_per_chan);
       }

       // The above will make sure that below I will or will not continue
       // inserting normal conn-switches as in normal VPR for
       // this track at this location, because once a connection in
       // Group1-6 or Group7 was detected, all 4 or 2 terminals
       // participate in hwrp.  Though this could be relaxed.
       // Next, at this time we are during delay lut's creation (done
       // only once at the begining of VPR run) then skip the whole stuff
       // related to hwrp and do it as in normal VPR:
       if ( during_lut_generation ) check_connection_type = NORMAL_SWITCH;
       // ......................................................................
       //
       if ( (check_connection_type == NORMAL_SWITCH)||	    
	    (check_connection_type == T2)|| 
	    (check_connection_type == T4) )
	 {
	   // Just let things "go" as in normal VPR if it's a normal switch
	   // OR T2, T4 because in this case the T2, T4 will be implemented
	   // as hard-wired but the fourth terminal will be connected
	   // "normally" otherwise would be dangling in the air and result 
	   // in rr_graph checking errors. So T's will be hard-wired T's
	   // with their fourt terminal connected normally.
	   // E.g.:
	   //         T2 
	   //      __|__
	   //     |  |  |
	   //   ----------- 
	   //     |__X__|
	   //        X    Fourth terminal of T2 connected normally.
	   //        X    in order not to remain in the air, unconnected
	   //             with no fanin or fanout.
	   *edge_list_ptr = insert_in_edge_list (*edge_list_ptr, to_node,
			   switch_types[0], &free_edge_list_head);
	   num_conn++;
	   if (switch_types[1] != OPEN) {
	     *edge_list_ptr = insert_in_edge_list (*edge_list_ptr, to_node,
                           switch_types[1], &free_edge_list_head);
	     num_conn++;
	   }
	 }
       // ......................................................................
       // Case hwrp's:  HORIZONTAL, VERTICAL, CROSS, L, T, etc.
       else {
	 //printf("%d",check_connection_type);
	 // Wherever I had a horizontal hwrp I also insert a vertical one.
	 // That is because if I used only two (the horizontal) terminals
	 // to embedd the H hwrp the other two may lead to routing segments
	 // (especially at margins) which will be isolated and rr_graph 
	 // checking routine will dicover that and exit.  There is no risk 
	 // of creating routing domains in this way because H hwrp's are not
	 // inserted allways at same track in all channels x.
	 if((check_connection_type == VERTICAL)||
	    (check_connection_type == T1)|| 
	    (check_connection_type == T3)) {
	   //printf("v");
	   //printf(" %d",from_track);
	   switch_types[0] = 1; // switch_hwrp
	   *edge_list_ptr = insert_in_edge_list (*edge_list_ptr, to_node,
				   switch_types[0], &free_edge_list_head);
	   num_conn++;
	 }

       }
       //
       // ......................................................................
    }//if
 }//for   
 
 return (num_conn);
}//eof
// ======================================================================= //



boolean is_sbox (int seg_num, int chan_num, int itrack, t_seg_details 
          *seg_details, boolean above_or_right) {

/* Returns TRUE if the specified segment has a switch box at the specified  *
 * location, FALSE otherwise.  The switch box is from the specified segment *
 * to the left (for chanx) or below (for chany) unless above_or_right is    *
 * TRUE.                                                                    */

 int seg_offset, start, length;
 boolean longline;
 
 length = seg_details[itrack].length;
 start = seg_details[itrack].start;
 longline = seg_details[itrack].longline;

/* NB: Periodicity is length for normal segments, length + 1 for long lines. */
 
 if (!longline) {
    seg_offset = (seg_num + chan_num - start + length) % length;
    seg_offset += above_or_right;    /* Add one if conn. is above or to right */
 }

 else {    /* Is a longline */
    seg_offset = (seg_num + chan_num - start + above_or_right) % (length + 1);
 }

 return (seg_details[itrack].sb[seg_offset]);
}
void get_switch_type (
	   boolean is_from_sbox, boolean is_to_sbox, 
	   short from_node_switch, short to_node_switch, 
	   short switch_types[2]) // Astia vor fi modificati aici.
                                  // oBs:  Can be hwrp_switch type, or
                                  // buffered or pass-transistir.

{
// This routine looks at whether the from_node and to_node want a switch,  
// and what type of switch is used to connect *to* each type of node       
// (from_node_switch and to_node_switch).  It decides what type of switch, 
// if any, should be used to go from from_node to to_node.  If no switch   
// should be inserted (i.e. no connection), it returns OPEN.  Its returned 
// values are in the switch_types array.  It needs to return an array      
// because one topology (a buffer in the forward direction and a pass      
// transistor in the backward direction) results in *two* switches.        

// cristinel.ababei
// Next two will be populated by this function.  This will convey
// the info about what kind of connection between the two tracks is
// existing inside the switch-box: in one direction, both directions,
// or no connection.
 switch_types[0] = OPEN;  /* No switch */
 switch_types[1] = OPEN;

 if (!is_from_sbox && !is_to_sbox) {  /* No connection wanted in either dir */
    switch_types[0] = OPEN;
 }

 else if (is_from_sbox && !is_to_sbox) {  /* Only forward connection wanted */
    switch_types[0] = to_node_switch;  /* Type of switch to go *to* to_node */
 }
 
 else if (!is_from_sbox && is_to_sbox) {

/* Only backward connection desired.  We're deciding whether or not to put *
 * in the forward connection.  Put it in if the backward connection uses   *
 * a bidirectional (pass transistor) switch.  Remember that the backward   *
 * connection uses a from_node_switch type of switch.                      */

// switch_inf is a global which has information about if the switch
// is buffered or is a simple pass-transistor, 
// R,C, etc.  This info is taken from .arch file
// inside read_arch()
    if (switch_inf[from_node_switch].buffered == FALSE) {
       switch_types[0] = from_node_switch;
    }
 }

 else {

/* Both a forward and a backward connection desired.  If the switch types   *
 * desired for the two connection are different, we have to reconcile them. */

    if (from_node_switch == to_node_switch) {
      // This should be the case for hwrp_switch type all the time.
      // hwrp_switch type will be bi-diretional and all the time
      // either buffered or simply continuous hard wire, but never
      // having in series any pass-transistor, because that is the idea
      // to get rid of pass-transistors and save area of SB!
      switch_types[0] = to_node_switch;
      // cristinel.ababei:  QWESTION:  If there are two connections desired
      // and both buffered why do not you also set switch_types[1]
      // in order to insert two edges? but you do that only on the
      // switch types are different?  Is this something that VPR's
      // developers missed or forgot?
      // I will temporarrily correct this for HWRP case only:
      /*---
      if (switch_inf[from_node_switch].HWRP_Type != NORMAL_SWITCH) {
          switch_types[1] = from_node_switch;
	  // So in this waythere will be two edges inserted.
       }
       ---*/
    }
    else {    /* Different switch types.  Reconcile. */
       if (switch_inf[to_node_switch].buffered) {
          switch_types[0] = to_node_switch;

          if (switch_inf[from_node_switch].buffered == FALSE) {

         /* Buffer in forward direction, pass transistor in backward.  Put *
          * in *two* edges.                                                */
              
             switch_types[1] = from_node_switch;
          }
       }
       else {   /* Forward connection is a pass transistor. */
          if (switch_inf[from_node_switch].buffered) {
             switch_types[0] = to_node_switch;
          }
          else {  

          /* Both forward and backward connections use pass transistors. *
           * use whichever one is larger, since you'll only physically   *
           * build one switch.                                           */
     
             if (switch_inf[to_node_switch].R < 
                            switch_inf[from_node_switch].R) {
                switch_types[0] = to_node_switch;
             }
             else if (switch_inf[from_node_switch].R < 
                            switch_inf[to_node_switch].R) {
                switch_types[0] = from_node_switch;
             }
             else {

             /* Two pass transistors have the same R, but are have different *
              * switch indices.  Use the one with lower index (arbitrarily), *
              * to ensure both switches are of the same type (since you can  *
              * only physically build one).  I'm being pretty dogmatic here. */

                if (to_node_switch < from_node_switch) {
                   switch_types[0] = to_node_switch;
                }
                else {
                   switch_types[0] = from_node_switch;
                }
             }
          }
       }
    }     /* End switch types are different */
 }   /* End both forward and backward connection desired. */
}//eof
// ======================================================================= //


