#include <stdio.h>
#include <math.h>
#include "util.h"
#include "vpr_types.h"
#include "globals.h"
#include "rr_graph_util.h"
#include "rr_graph.h"
#include "rr_graph2.h"
#include "rr_graph3.h"
#include "rr_graph_sbox.h"
#include "check_rr_graph.h"
#include "rr_graph_timing_params.h"
#include "rr_graph_indexed_data.h"
#include <assert.h>
#include<stdlib.h>
#include<time.h>


// =================== Some globals ====================================== //
// Variables below are the global variables used for hard-wired routing 
// patterns stuff.

// Used initially for counting overall - physically existing in the arch of
// the FPGA - number of connctions.
int  Num_Group123456_type_connections;
int  Num_Group7_type_connections;
// These next fractions should be taken (found out) thru analysis
// by Eli and Ryan.
int fraction_H, fraction_V, fraction_CROSS;
int fraction_L; // L1 L2 L3 and L4
int fraction_T; // T1 T2 T3 and T4
int fraction_hwrp; // Set by user.
int NG, N_H, N_V, N_CROSS, N_T, N_L; 
int Step_H, Step_V, Step_HV, Step_CROSS, Step_T, Step_L; 
int Num_hwrp, Step_hwrp;

// Next is a global array which "emulates" the array of all
// switch-blocks 1..nx-1, 1..ny-1 (ie all those which have four clb 
// neighbors) because only at the level of these SBs' I will insert 
// hwrp's.  This array will be created and populated with appropriate
// distributions of hwrp's when "virtual" functions will be used.
// Every (i,j) location will contain info about all terminals
// on the four margins of the switch-block (a margin has a number of 
// CW terminals).
int *** physical_array_switch_boxes;

// For leakage-power calculations:
int   counter_conn_buffers[2];
int   counter_conn_pass_tran[2];
int   total_hwrp_switch_connections;
int   total_efective_eliminated_due_to_hwrp_insertion;

// ========================== Subroutines local to this module =========== //

void generate_distributions_hwrp(	   
		 int nodes_per_chan,
		 t_seg_details *seg_details_x, 
		 t_seg_details *seg_details_y );
boolean is_it_ok_to_insert_hwrp_at_this_location(
			     int i,int j,int k,
			     int nodes_per_chan,
			     t_seg_details *seg_details_x, 
			     t_seg_details *seg_details_y );
void dump_physical_array_switch_boxes(char *file_name, int nodes_per_chan);

// ======================================================================= //



void virtual_alloc_and_load_rr_graph (
	    int **rr_node_indices, 
	    int ***clb_opin_to_tracks, 
	    struct s_ivec **tracks_to_clb_ipin,
	    int **pads_to_tracks, 
	    struct s_ivec *tracks_to_pads, 
	    int Fc_output, int Fc_input, int Fc_pad, 
	    int nodes_per_chan, 
	    enum e_route_type route_type, 
	    struct s_det_routing_arch det_routing_arch, 
	    t_seg_details *seg_details_x, t_seg_details *seg_details_y)
{
 // cristinel.ababei:  This is for hwrp's stuff.  This way, in order 
 // to make use of VPR coding style as much as possible to make my job easier. 
 // Does almost same things as alloc_and_load_rr_graph() (that's why it's
 // called "virtual") but
 // with the sole goal of counting all connections that will
 // appear inside switch-boxes.  These connections can be
 // made between 4 different routing-segments ending at a SB or between
 // 2 routing-segments which cross the SB but still require connection.
 // Routing-segments at their turn can be of type 1,2,6,long.
 // I need to know a-priori these numbers of connections such that
 // when the actual rr_graph will be constructed to know which
 // ones (only a fraction of them) to be made of type 
 // T or L or + or -- or |.
 int i=0, j=0, k=0;
 enum e_switch_block_type switch_block_type;
 int delayless_switch, wire_to_ipin_switch, num_segment;


 switch_block_type = det_routing_arch.switch_block_type;
 delayless_switch = det_routing_arch.delayless_switch;// It's an index.
 wire_to_ipin_switch = det_routing_arch.wire_to_ipin_switch;// It's an index.
 num_segment = det_routing_arch.num_segment;


 // Reset global counters.
 Num_Group123456_type_connections = 0;
 Num_Group7_type_connections = 0;
 // Allocate "physical_array_switch_boxes" which will have to contain
 // info about the distribution of hwrp's.  Then, reset all its elements
 // to -1 which means this terminal of the SB is not part of a
 // switch-connection (is sort of open in the sense that these tracks
 // do not require switch-block connections here).
 //       3  4  5
 //     __|__|__|__
 // 2__|           |__8
 // 1__|    SB     |__7
 // 0__|           |__6
 //    |___________|
 //       |  |  |
 //       9  10 11 
 // physical_array_switch_boxes[i][j] = 0 1 2 3 4 5 6 7 8 9 10 11
 // So, when a terminal will be set as participating to a SB connection
 // it will be discovered as 0 1 2 but all their counterparts (say 3,6,9
 // in case of 0 for disjoint sb style) will be marked.

 physical_array_switch_boxes = (int ***) alloc_matrix3 (
		    0,nx, 0,ny, 0,(4*nodes_per_chan), sizeof(int));
 for (i=0;i<=nx;i++) { 
    for (j=0;j<=ny;j++) {
      for (k=0;k<=(4*nodes_per_chan);k++) {
	physical_array_switch_boxes[i][j][k] = -1;
      }
    }
 }


 // Sweep all locations in order to count all connections that will
 // appear inside switch-boxes.
 for (i=1; i<=nx; i++) {
   for (j=1; j<=ny; j++) {

     if (clb[i][j].type == CLB) {
       // Next, I am keeping - with the risk of redundancy - all
       // arguments in order to simplify coding.
       virtual_build_rr_xchan (rr_node_indices, route_type, tracks_to_clb_ipin, 
				 tracks_to_pads, i, j, nodes_per_chan, switch_block_type, 
				 wire_to_ipin_switch, seg_details_x, seg_details_y,
				 CHANX_COST_INDEX_START);
     }
     else if (clb[i][j].type == IO) {
       continue;
     }
     else if (clb[i][j].type != ILLEGAL) {
       printf ("Error.  Block at (%d, %d) has unknown type (%d).\n", i, j, 
	       clb[i][j].type);
       exit (1);
     }//if
     
   }//for
 }//for
 

 // Counting results.
 printf ("Total number of physically existing connections "
	 "of type Group123456 = %d.\n", Num_Group123456_type_connections);
 printf ("Total number of physically existing connections "
	 "of type Group7 = %d.\n", Num_Group7_type_connections);
 printf ( "This means an average of about %g Group123456 and %g Group7 "
	  "connections on Switch-Block.\n", 
	  (double)Num_Group123456_type_connections/(double)((nx-1)*(ny-1)),
	  (double)Num_Group7_type_connections/(double)((nx-1)*(ny-1)) );
 printf ("oBs:  A fraction of these will be made hard-wired routing patterns.\n");




 // Until now I practically only counted all switch-connections.
 // If "hwrp"-stuff is disabled (i.e., normal-usual VPR is required to
 // be run) then, this counting I will use only to count total number
 // of buffers and their pass transitors inside all SB's, which 
 // are sorrounded by CLB's. Attention here: this count will not reflect
 // those bufferes and pass-tran inside SB's between channels x,y and IO
 // channels!
 if (disable_hwrp == TRUE) {
   // This is pure VPR where all switch-connections are "usual" with
   // no hwrp's inserted.
   total_hwrp_switch_connections = 0;
   total_efective_eliminated_due_to_hwrp_insertion = 0;

   // So let's count all buffers and pas transistors because we do not call
   // generate_distributions_hwrp() anymore in this case. This counting is 
   // needed in power.c for power calculations.
   for (i=0;i<2;i++) { 
     // These are globals, in order to avoid passing thru too many arguments
     // to different functions.
     counter_conn_buffers[i] = 0;
     counter_conn_pass_tran[i] = 0;
   }
   // Let's see how many switch-connections we have and how they determine the
   // total number of bufferes and pass transitors inside switch-blocks. 
   // i=0 is (i) usual switch-connection
   counter_conn_buffers[0] += 
     4 * Num_Group123456_type_connections + 
     // 4 because each segment-ending entering the SB connects to other three
     // on the other three sides and there is buffer sharing.
     2 * Num_Group7_type_connections;
   counter_conn_pass_tran[0] += 
     12 * Num_Group123456_type_connections + 
     2 * Num_Group7_type_connections;
 }
 else {
   // Now that counting is over, generate distributions of hwrp's.
   generate_distributions_hwrp(nodes_per_chan,seg_details_x,seg_details_y);
   // Inside the above function counting of all buffers and pass-trans 
   // will be done for both kinds of switch-connections: (i) usual and 
   // (ii) hwrp
 }


}//eof
// ======================================================================= //



void virtual_build_rr_xchan (
	     int **rr_node_indices, 
	     enum e_route_type route_type, 
	     struct s_ivec **tracks_to_clb_ipin, 
	     struct s_ivec * tracks_to_pads, 
	     int i, int j, int nodes_per_chan, 
	     enum e_switch_block_type switch_block_type, 
	     int wire_to_ipin_switch, 
	     t_seg_details *seg_details_x, t_seg_details *seg_details_y,
	     int cost_index_offset)
{
 // cristinel.ababei
 // oBs: Developed starting from build_rr_xchan().
 // Used for sweeping all tracks in all channels x with the goal of
 // counting overall number of Group123456_type_connection
 // or Group7_type_connection in the entire FPGA.
 // Called for all i,j pairs.
 int itrack, inode, istart, iend, length, seg_index;
 int num_edges=0;

 for (itrack=0;itrack<nodes_per_chan;itrack++) {

   istart = get_closest_seg_start (seg_details_x, itrack, i, j);
   if (istart != i)
     continue;        // Not the start of this segment.

   iend = get_seg_end (seg_details_x, itrack, istart, j, nx);


   if (j == 0) {                   // Between bottom pads and clbs.
     continue;
   }
   else if (j == ny) {             // Between top clbs and pads. 
     continue;
   }
   else {                          // Between two rows of clbs.
     // Channels above and below both exist.
     num_edges += virtual_get_xtrack_to_ytracks (
		   istart, iend, j, 
		   itrack, j+1,
		   nodes_per_chan, 
		   rr_node_indices, 
		   seg_details_x,  seg_details_y, 
		   switch_block_type); 

     num_edges += virtual_get_xtrack_to_ytracks (
		   istart, iend, j, 
		   itrack, j,
		   nodes_per_chan, 
		   rr_node_indices, 
		   seg_details_x,  seg_details_y, 
		   switch_block_type); 

   }
    
   if (istart != 1) {     // x-chan to left exists.
     continue;
   }
   if (iend != nx) {      // x-chan to right exists.
     continue;
   }

 }// for (itrack=0;

}//eof
// ======================================================================= //



int virtual_get_xtrack_to_ytracks (
	     int from_istart, int from_iend, int from_j, 
	     int from_track, int to_j, 
	     int nodes_per_chan,
	     int **rr_node_indices, 
	     t_seg_details *seg_details_x, t_seg_details *seg_details_y, 
	     enum e_switch_block_type switch_block_type)
{
 // cristinel.ababei
 // oBs: Developed starting from get_xtrack_to_ytracks().
 // Used for counting overall number of Group123456_type_connection
 // or Group7_type_connection in the entire FPGA.  Later then some
 // fraction of these connections will be made HWRP; but first
 // I have to see how many are in total because the number of connections
 // and their type is not really same in every SB and also CW may be small
 // enough and not have enough connections inside a SB to cover all
 // possible patterns L,T,+,|,--.
 int to_node=0, i=0, to_track=0, iconn=0;
 int from_node_switch, to_node_switch;
 int num_conn=0;
 short switch_types[2];
 boolean is_x_sbox, is_y_sbox, yconn_to_above;
 struct s_ivec conn_tracks;



 // I skip this case to avoid overcounting.
 // In other words, looking at a SB:
 //   ___
 // 1|/\2|
 //  |\ \| 
 //  |4\/|3
 //   ---    I will detect "4" only to count the whole connection
 if (to_j <= from_j) {
   yconn_to_above = TRUE;  // The connection goes up from ychan to xchan.
   // In timpul acestui counting ma uit doar la conexiuni x-->y
   // inspre jos (left-->right,bottom) in interiorul SB si numai la dreapta.
   // Asta e ptr. a evita overcounting of conections.
 }
 else {
   yconn_to_above = FALSE;
   return 0; 
 }


 from_node_switch = seg_details_x[from_track].wire_switch;

 for (i=from_istart;i<=from_iend;i++) {

   // Acum mergem de-a lungul wire-segmentului de la "from_start"
   // till "from_end" si for every location ma uit to see whether 
   // there should be a switch "on" this segment and if so I have to "look" at 
   // possible connections to tracks (wire-segments) in y channels.


   // ===> Diagonal connection to left (from xchan to ychan)
   // -- TO LEFT
   // I skip this case because I will look at only "to right"
   // That is because once I dicovered that there is a connection 
   // to right, I know that I found a Group123456_type_connection
   // or Group7_type_connection anyway and I have to count it 
   // only once.
   //
   // ---------       ---------
   // SB      |_seg___|      SB
   //        /|       |\             And we go from smaller i to larger i
   //       / |       | \            sweeping the roting segment from "from_istart"
   //                                till "from_iend"
   //  LEFT      :-)      RIGHT
   // (i-1,j)   (i,j)     (i,j)
   //

   // ===>  Diagonal connection to right (from xchan to ychan) 
   // -- TO RIGHT
   is_x_sbox = is_sbox (i, from_j, from_track, seg_details_x, TRUE);
   conn_tracks = get_switch_box_tracks (
		i, from_j,  // Loc. on CHANX during mersului.
		from_track, // Inside xchan
		CHANX, 
		i, to_j,    // Which y-chan location
		CHANY, 
		switch_block_type, nodes_per_chan);
  
   for (iconn=0;iconn<conn_tracks.nelem;iconn++) {
     to_track = conn_tracks.list[iconn];
     is_y_sbox = is_sbox (to_j, i, to_track, seg_details_y, yconn_to_above);
     to_node_switch = seg_details_y[to_track].wire_switch;
     
     get_switch_type (is_x_sbox, is_y_sbox, from_node_switch, to_node_switch,
		      switch_types);

     if (switch_types[0] != OPEN) {
       if ( i==from_iend ) {
	 // If currently sweeped location is one of the
	 // endings of the routing-segment under scrutinity 
	 // and it has a connection; then this connection can only 
	 // be one of the type between four endings entering same
	 // switch-block (assuming disjoint SB style)
	 Num_Group123456_type_connections = 
	   Num_Group123456_type_connections + 1;
	 num_conn++;
	 // Now mark all four terminals of this SB at this track 
	 // as terminals participating to a SB connection where all terminals
	 // end at this SB.
	 physical_array_switch_boxes[i][to_j][from_track] = 1;
	 physical_array_switch_boxes[i][to_j][from_track + 1*nodes_per_chan ] = 1;
	 physical_array_switch_boxes[i][to_j][from_track + 2*nodes_per_chan ] = 1;
	 physical_array_switch_boxes[i][to_j][from_track + 3*nodes_per_chan ] = 1;
       }// if
       else {
	 // This case it's a connection between middle points
	 // (or other points alongs segments but different from its
	 // endings) and this envolves only one single connection
	 // cross-like (type 7) when segments overlap
	 Num_Group7_type_connections = 
	   Num_Group7_type_connections + 1;
	 /*---
	 num_conn++;
	 // Now mark all four terminals of this SB at this track 
	 // as terminals participating to a SB connection where all terminals
	 // are terminals due to two routing segments which overlap
	 // and require connection type 7 at this SB location.

	 // For a first implementation I do not use this connection
	 // for implementing hwrp's.

	 physical_array_switch_boxes[i][to_j][from_track] = 2;
	 physical_array_switch_boxes[i][to_j][from_track + 1*nodes_per_chan ] = 2;
	 physical_array_switch_boxes[i][to_j][from_track + 2*nodes_per_chan ] = 2;
	 physical_array_switch_boxes[i][to_j][from_track + 3*nodes_per_chan ] = 2;
	 ---*/
       }
     }//if
   }//for

 }// for (i=from_istart;i<=from_iend;i++) ---> End for length of segment.
 
 return num_conn;

}//eof
// ======================================================================= //


// oBs:  This is a local function.
void generate_distributions_hwrp(	   
		 int nodes_per_chan,
		 t_seg_details *seg_details_x, 
		 t_seg_details *seg_details_y )
{
 // Called after the counting of all switch-connections (using above
 // "virtual" functions) was done.  Here we populate 
 // "physical_array_switch_boxes" which will have to contain
 // info about the distribution of hwrp's.  This is done based on the
 // percentage - set by user - of hwrp's and based on the fractions 
 // of each type of hwrp as found out by Eli's analysis. Important:
 // is that the above "virtual" functions "marked" already  all terminals
 // which participate in switch-connections with "1" and "2".  
 // Now we only have to set 
 // a percentage of those to be of one type of hwrp or another. 

 int i=0, j=0, k=0;
 int local_counter_group123456=0;
 int local_counter_group7=0;
 int local_counter_both_groups=0;
 int  local_counter_of_inserted_hwrp=0;
 int  make_L_lt_br=1;
 int  make_T__l_t_b_r=1;
 int seed1;
 
 // seed1=time(NULL);
 // Compute number of each hwr-pattern using the percentages 
 // (fractions) found out with Eli and Ryan's routing analysis and/or
 // with my statistical analysis.
 // 
 //        __________|____SB
 //       |        __|__  |
 //    ___|__| T1    T2   |
 //       |  |          |_|___T3
 //       |             | |
 //       |  |      L2|   |
 //    ___|__| L1     |___|___
 //       |        _______|___  
 //    ___|__     |  ___  |  L3
 //       |__|____|___|___|
 //          |L4  |   |T4


 fraction_H = 20;
 fraction_V = 20;
 fraction_CROSS = 10;
 fraction_L = 30; // L1 L2 L3 and L4
 fraction_T = 20; // T1 T2 T3 and T4
 // This means that if I look at groups of 10 candidate connections 
 // (ie which can be used to implement hwrp's) then, roughly
 // 2 of them will be H, 2 will be V, 1 CROSS, 3 L and 2 T
 // So, for the moment I am doing this manually:  I am creating
 // this array of ten (10) elements which will be used during
 // the distribution of hwrps among all possible/candidates
 //The following block of code distributes the patterns according to the
 //lengths of the patterns
 int index_man_distr=0;
 int code_array[5]={10,20,30,41,51};
 int frac_patterns[3][5]={{20,20,10,30,20},{20,20,10,30,20},{20,20,10,30,20}};
//int frac_patterns[3][5]={{1,15,10,45,15},{24,24,10,32,10},{19,19,10,38,14}};
 static int manual_distribution[3][100];
 static int array_init_flag=0;
 int length=0;
 if(!array_init_flag)
   {
     for(length=0;length<3;length++)
      { 
         for(i=0;i<100;i++)
           manual_distribution[length][i]=-1;
         int code_ctr=0;int t,dis_flag;
         for(i=0;i<5;i++)
           {
               for(j=0;j<frac_patterns[length][i];j++)
                 {do{
                      dis_flag=1;
                      while(100<=(t=rand()/(RAND_MAX/100)));
                      if(manual_distribution[length][t]==-1)
                         dis_flag=0;
                     }
                  while(dis_flag==1);
		 manual_distribution[length][t]=code_array[i];
                 }
	   }  
      } 
   }
 array_init_flag=1;  

 //Satish.
 //I have made the manual_distribution array static and I have another flag- array_init_flag.
 //This is because I want to populate the distribution array only once according to the
 //percentages. If i keep doing this everytime, then the percentages will change and
 //when the tool checks for routing consistency after it finishes routing, it will fail
 //since it will insert harps in different locations while checking and that will fail the
 //test. It is for this reason that i cannot make the orientations of the patterns random. 
 //making orientations also random requires more thinking
 
/*static int manual_distribution[100];
 //static int array_init_flag=0;
  if(!array_init_flag)
   { 
  for (i=0;i<100;i++)
    manual_distribution[i]=-1;
  int t=0;int dis_flag;
  //srand(seed1);
   for(i=0;i<fraction_H;i++)
    {do{
       dis_flag=1;
       while(100<=(t=rand()/(RAND_MAX/100)));
       if(manual_distribution[t]==-1)
         dis_flag=0;
       }
     while(dis_flag==1);
    manual_distribution[t]=10;
    }

    for(i=0;i<fraction_V;i++)
    {do{
      dis_flag=1;
      while(100<=(t=rand()/(RAND_MAX/100)));
      if(manual_distribution[t]==-1)
        dis_flag=0;
       }
    while(dis_flag==1);
   manual_distribution[t]=20;
   }

   for(i=0;i<fraction_CROSS;i++)
   {do{
     dis_flag=1;
      while(100<=(t=rand()/(RAND_MAX/100)));
     if(manual_distribution[t]==-1)
        dis_flag=0;
       }
     while(dis_flag==1);
    manual_distribution[t]=30;
     }
    for(i=0;i<fraction_L;i++)
     {do{
       dis_flag=1;
       while(100<=(t=rand()/(RAND_MAX/100)));
      if(manual_distribution[t]==-1)
        dis_flag=0;
       }
   while(dis_flag==1);
   manual_distribution[t]=41;
}
    
  for(i=0;i<fraction_T;i++)
   {do{
     dis_flag=1;
     while(100<=(t=rand()/(RAND_MAX/100)));
     if(manual_distribution[t]==-1)
       dis_flag=0;
      }
   while(dis_flag==1);
   manual_distribution[t]=51;
   }
   }
   array_init_flag=1;*/
 // int manual_distribution[10]={41, 30, 51, 51, 10, 20, 20, 41,10,41};
 //int manual_distribution[10] =
 //  { 30, 20, 41, 10, 51, 41, 20, 51, 10, 41 };
 //  C   V   L   H   T   L   V   T   H   L  : codes
 //  1   2   3   4   5   6   7   8   9   10 : indices
 // 
 // Observe that above all patterns are "distributed" some sort of 
 // equidistantly among the ten indices such that their fractions is
 // correct. For example we have "H" on locations 4,9, which makes 
 // two out of ten (ie 20%).
 // The above distribution is done using following "mapping" of 
 // connection types to numerical codes:
 // 10 20 30  41 42 43 44  51 52 53 54  60 70
 // H  V  C   L1 L2 L3 L4  T1 T2 T3 T4  NS NSAA


 // Also the total fraction (out of all connections) which we
 // want as hard-wired is set by user and for now let's make it:
 fraction_hwrp = 100;
 // For the moment i put it 100 but in fact because many of them will
 // be skiped in order to avoid adjacent hwrp's, the final fraction
 // of inserted hwrp's will be about 50%. How to solve this more
 // elegantly requires more thinking.


 // -------------------------------------------------------------------------
 // 1 --- Pun pe hold situatia when I consider for hwrps also the Group7
 // connections.
 /*---
 // Total number of connections:
 NG = Num_Group123456_type_connections + Num_Group7_type_connections;
 // Number of Horizontal,Vertical connections and "step" used while counting
 // and setting them to this type.
 N_H = (int)( ((double)fraction_H /100)*((double)fraction_hwrp /100)*(double)NG );
 //printf ("N_H = %d.\n", N_H);
 N_V = (int)( ((double)fraction_V /100)*((double)fraction_hwrp /100)*(double)NG );
 Step_H = (int)( (double)Num_Group123456_type_connections /(double)N_H );
 Step_V = (int)( (double)Num_Group123456_type_connections /(double)N_V );
 Step_HV = (int)( (double)Num_Group123456_type_connections /(double)(2*N_H) );
 // Which means that H and V connections will have to be made only
 // using connections from group Group123456.
 printf ("Step_H == Step_V= %d, ", Step_H);
 printf ("Step_HV= %d, ", Step_HV);
 // Number of CROSS connections and "step" used while counting
 // and setting them to this type.
 N_CROSS = (int)( ((double)fraction_CROSS/100)*
		  ((double)fraction_hwrp/100)*(double)NG );
 Step_CROSS = (int)( (double)NG / (double)N_CROSS );
 printf ("Step_CROSS = %d, ", Step_CROSS); 
 // Used NG because cross-type connections
 // can be implemented with both  Group123456 and  Group7 type of connects.
 // Number of T,L connections and "step" used while counting
 // and setting them to this type.
 N_T = (int)( ((double)fraction_T /100)*
	      ((double)fraction_hwrp /100)*(double)NG );
 Step_T = (int)( (double)NG / (double)N_T );
 printf ("Step_T = %d, ", Step_T); 
 N_L = (int)( ((double)fraction_L / 100)*
	      ((double)fraction_hwrp / 100)*(double)NG );
 Step_L = (int)( (double)NG / (double)N_L ); 
 printf ("Step_L = %d.\n", Step_L);
 ---*/
 // -------------------------------------------------------------------------

 // -------------------------------------------------------------------------
 // 2 --- For now I will use only Group123456 connections
 // to implement hwrps.
 /*---
 // Total number of connections:
 NG = Num_Group123456_type_connections;
 // Number of Horizontal,Vertical connections and "step" used while counting
 // and setting them to this type.
 N_H = (int)( ((double)fraction_H /100)*((double)fraction_hwrp /100)*(double)NG );
 //printf ("N_H = %d.\n", N_H);
 N_V = (int)( ((double)fraction_V /100)*((double)fraction_hwrp /100)*(double)NG );
 Step_H = (int)( (double)NG /(double)N_H );
 Step_V = (int)( (double)NG /(double)N_V );
 Step_HV = (int)( (double)NG /(double)(2*N_H) );
 printf ("Step_H == Step_V= %d, ", Step_H);
 printf ("Step_HV= %d, ", Step_HV);
 // Number of CROSS connections and "step" used while counting
 // and setting them to this type.
 N_CROSS = (int)( ((double)fraction_CROSS/100)*
		  ((double)fraction_hwrp/100)*(double)NG );
 Step_CROSS = (int)( (double)NG / (double)N_CROSS );
 printf ("Step_CROSS = %d, ", Step_CROSS); 
 // Used NG because cross-type connections
 // can be implemented with both  Group123456 and  Group7 type of connects.
 // Number of T,L connections and "step" used while counting
 // and setting them to this type.
 N_T = (int)( ((double)fraction_T /100)*
	      ((double)fraction_hwrp /100)*(double)NG );
 Step_T = (int)( (double)NG / (double)N_T );
 printf ("Step_T = %d, ", Step_T); 
 N_L = (int)( ((double)fraction_L / 100)*
	      ((double)fraction_hwrp / 100)*(double)NG );
 Step_L = (int)( (double)NG / (double)N_L ); 
 printf ("Step_L = %d.\n", Step_L);
 ---*/
 // -------------------------------------------------------------------------

 // -------------------------------------------------------------------------
 // --- 3
 Num_hwrp = (int)( ((double)fraction_hwrp /100)*
		   (double)Num_Group123456_type_connections );
 Step_hwrp = (int)( (double)Num_Group123456_type_connections /(double)Num_hwrp );
 printf("Insertion of hwrp's step =%d ",Step_hwrp);
 printf("in order to meet the required fraction of %d [1/100] hwrp's ",fraction_hwrp);
 printf("(which is %d) out of %d\n",Num_hwrp, Num_Group123456_type_connections);
 // -------------------------------------------------------------------------

 // -------------------------------------------------------------------------
 // Stuff related to power calculation:  Here, because I insert actually
 // hwrp's, I have to count all switch-connections of each type. For now 
 // inside SBs there are two kinds of connections: (i) usual (in normal VPR)
 // which are of type-1 or type-2
 // type-1 is when all four endings of four segments participate to 6 connections
 // inside the SB, and type-2 when there is only one conection of two overlaping 
 // longer segments.
 // (ii) hwrp's connections which can be of differet L,T,H,V,Cross types.
 // Therefore, I have to actually count total number of buffered (shared) 
 // and total number of pass transistors. This info will be used in power.c
 // for leakage power calculation.
 //
 for (i=0;i<2;i++) { 
   // These are globals, in order to avoid passing thru too many arguments
   // to different functions.
   counter_conn_buffers[i] = 0;
   counter_conn_pass_tran[i] = 0;
   // i=0 is (i) usual 
   // i=1 is (ii) hwrp
   // In the future for complete flxibility "i" will have to go up to the
   // number of different switch-connections. But for now (and I guess nobody
   // will care introducing different types) we have only one switch (described
   // inside .arch file) which is the usual and a second one introduced by
   // me to emulate (ii) hwrp. Their indices are 0 and 1 when "looking" at
   // the globals switch_inf[].
 }


 // -------------------------------------------------------------------------


 // ---> Let's set V and H connections.
 local_counter_group123456 = 0; // Reset.
 local_counter_of_inserted_hwrp = 0; // Reset.
 index_man_distr=0; // Reset.
 make_L_lt_br = 1; // Reset.
 make_T__l_t_b_r = 1; // Reset.
 int local_harp_counter[3]={0,0,0};
 int length_index=0;
 // Counters:
 int  hwrp_H_counter = 0, hwrp_V_counter = 0;
 int  hwrp_L_counter = 0, hwrp_T_counter = 0;
 int  hwrp_CROSS_counter = 0;
 //Satish. Variable to increase the percentage of hwrps
 int false_counter =0;
 boolean is_it_ok_to_insert=TRUE;
 int Lrot=2;
 int Lpattern;
 int seed;
 //seed=time(NULL);
 //srand(seed);
 for (i=1;i<nx;i++) { 
   for (j=1;j<ny;j++) {
     for (k=0;k<nodes_per_chan;k++) {


       // oBs:  Important aspect:  In order not to have hwrp's connected
       // among tehm I have to make sure that I do not insert adjacent
       // hwrps.  Otherwise I would have connected "trees" of hwrps
       // with dangling branches of them possibly not beeing used
       // and adding extra capacitance and in this way impairing the goal
       // of hwrps.
       if (physical_array_switch_boxes[i][j][k] == 1) {
	 // This is connection from group Group123456.
	 local_counter_group123456 = local_counter_group123456 + 1;
	 //Satish. To increase the percentage of hwrps, i need to allow
       //certain hwrps to connect to each other. Else, i will never get 
        //more than 50%
	 is_it_ok_to_insert = 
	   is_it_ok_to_insert_hwrp_at_this_location(i,j,k,nodes_per_chan,
					     seg_details_x,seg_details_y);
	 //is_it_ok_to_insert=TRUE;//Satish. This will make all the connections hard-wired
        
         //Satish. I can change FALSE to TRUE and vice versa and this will
         //reduce the number of hard-wired patterns to a number below 50% i.e.
         //I make some of the locations where i can insert hwrps as false locations

         if(is_it_ok_to_insert==TRUE)
	    false_counter++;
         //for 55% use 3 and 7. for 70% use 2,5 and 9
	 if(!(false_counter%3)||!(false_counter%7)||!(false_counter%4))
	   is_it_ok_to_insert=FALSE;
         if(false_counter==15)
	    false_counter=0;
	 //What the above block of code does is, to count the number of false locations
        //then it makes 3 out of 10 false locations true. hopefully this will increase
       //the number of hwrps by 15 % ( 3/10 of 50 % approximately)
        if (is_it_ok_to_insert == FALSE) {
	  // Do nothing. Skip.
	  //printf("%d",is_it_ok_to_insert);
	}
	else { // TRUE
	 //printf("%d",is_it_ok_to_insert);
	 // If it's multiple of Step_H then make this conenction
	 // hwrp of type Horizontal by marking all 4 terminals
	 // appropriately.
	 if ( (local_counter_group123456 % Step_hwrp) == 0 ) {
	   // I have to make it hwrp. For that I use 
	   // manual_distribution[10] and insert the hwrp which is
	   // in the "pipeline" of ten.
	   // manual_distribution[10] =
	   // { 30, 20, 41, 10, 51, 41, 20, 51, 10, 41 };
	   //   C   V   L   H   T   L   V   T   H   L  : codes
	   //   1   2   3   4   5   6   7   8   9   10 : indices
	   //   and information about mapping:
	   // 10 20 30  41 42 43 44  51 52 53 54  60 70
	   // H  V  C   L1 L2 L3 L4  T1 T2 T3 T4  NS NSAA
	   
           //Satish. Here I find out what the length of the segment is and
           //use the appropriate percentages of HARPs
	   if(seg_details_x[k].length==1)
	     length_index=0;
           else if (seg_details_x[k].length==2)
             length_index=1;
           else if (seg_details_x[k].length==6)
             length_index=2; 
           else 
             length_index=0;
	   //index_man_distr = 
	   //  local_counter_of_inserted_hwrp % 100; // modulo 10. 
	   index_man_distr=local_harp_counter[length_index]%100;
           // while(10<=(index_man_distr=rand()/(RAND_MAX/10)));
	     local_counter_of_inserted_hwrp = 
	     local_counter_of_inserted_hwrp  + 1;
	     local_harp_counter[length_index]++;

	   if ( (manual_distribution[length_index][index_man_distr] == 10)||
		(manual_distribution[length_index][index_man_distr] == 20) ) { // H or V
	     hwrp_H_counter++;
	     hwrp_V_counter++;
	     physical_array_switch_boxes[i][j][k] = 10;
	     physical_array_switch_boxes[i][j][k + 1*nodes_per_chan ] = 20;
	     physical_array_switch_boxes[i][j][k + 2*nodes_per_chan ] = 10;
	     physical_array_switch_boxes[i][j][k + 3*nodes_per_chan ] = 20;
	   }
	   else if ( manual_distribution[length_index][index_man_distr] == 30) { // CROSS
	     hwrp_CROSS_counter++;
	     physical_array_switch_boxes[i][j][k] = 30;
	     physical_array_switch_boxes[i][j][k + 1*nodes_per_chan ] = 30;
	     physical_array_switch_boxes[i][j][k + 2*nodes_per_chan ] = 30;
	     physical_array_switch_boxes[i][j][k + 3*nodes_per_chan ] = 30;
	   }
	   else if ( manual_distribution[length_index][index_man_distr] == 41) { // L
	     hwrp_L_counter++;
	     //while(Lrot<=(Lpattern=rand()/(RAND_MAX/Lrot)));
	      //printf("\nLpattern = %d",Lpattern);
	      //make_L_lt_br=Lpattern;
             //if(Lpattern==0){
             if ( make_L_lt_br == 1 ) { // left-top and bottom-right
	       //    |L1
	       //  __|  ___
	       //      |   L3
	       physical_array_switch_boxes[i][j][k] = 41;
	       physical_array_switch_boxes[i][j][k + 1*nodes_per_chan ] = 41;
	       physical_array_switch_boxes[i][j][k + 2*nodes_per_chan ] = 43;
	       physical_array_switch_boxes[i][j][k + 3*nodes_per_chan ] = 43;
	       make_L_lt_br = 0;
	     }
	     else {                     // left-bottom and top-right
	       //       |L2
	       //   __  |___
	       // L4  |
	       physical_array_switch_boxes[i][j][k] = 44;
	       physical_array_switch_boxes[i][j][k + 1*nodes_per_chan ] = 42;
	       physical_array_switch_boxes[i][j][k + 2*nodes_per_chan ] = 42;
	       physical_array_switch_boxes[i][j][k + 3*nodes_per_chan ] = 44;
	       make_L_lt_br = 1;
	     }
	   }
	   else if ( manual_distribution[length_index][index_man_distr] == 51) { // T
	     hwrp_T_counter++;
             Lrot=4;
	     // while(Lrot<=(Lpattern=rand()/(RAND_MAX/Lrot)));
             //make_T__l_t_b_r=Lpattern;
	     if ( make_T__l_t_b_r == 0 ) {
	       physical_array_switch_boxes[i][j][k] = 51;
	       physical_array_switch_boxes[i][j][k + 1*nodes_per_chan ] = 51;
	       physical_array_switch_boxes[i][j][k + 2*nodes_per_chan ] = 51;
	       physical_array_switch_boxes[i][j][k + 3*nodes_per_chan ] = 51;
	       make_T__l_t_b_r = 2;
	     }
	     else if ( make_T__l_t_b_r == 1 ) {
	       physical_array_switch_boxes[i][j][k] = 52;
	       physical_array_switch_boxes[i][j][k + 1*nodes_per_chan ] = 52;
	       physical_array_switch_boxes[i][j][k + 2*nodes_per_chan ] = 52;
	       physical_array_switch_boxes[i][j][k + 3*nodes_per_chan ] = 52;
	       make_T__l_t_b_r = 3;
	     }
	     else if ( make_T__l_t_b_r == 2 ) {
	       physical_array_switch_boxes[i][j][k] = 53;
	       physical_array_switch_boxes[i][j][k + 1*nodes_per_chan ] = 53;
	       physical_array_switch_boxes[i][j][k + 2*nodes_per_chan ] = 53;
	       physical_array_switch_boxes[i][j][k + 3*nodes_per_chan ] = 53;
	       make_T__l_t_b_r = 4;
	     }
	     else {
	       physical_array_switch_boxes[i][j][k] = 54;
	       physical_array_switch_boxes[i][j][k + 1*nodes_per_chan ] = 54;
	       physical_array_switch_boxes[i][j][k + 2*nodes_per_chan ] = 54;
	       physical_array_switch_boxes[i][j][k + 3*nodes_per_chan ] = 54;
	       make_T__l_t_b_r = 1;
	     }
	   }	     
	 }//if ( (local_counter_group123456 % Step_hwrp) == 0 )
	}
       }//if ( physical_array_switch_boxes[i][j][k] == 1 )
      
 
     }//for
   }//for
 }//for

 printf ("Distribution of inserted hard-wired-routing-patterns:\n"
	 " H, V = %d\n L = %d\n T = %d\n CROSS = %d\n", hwrp_H_counter,
	 hwrp_L_counter, hwrp_T_counter, hwrp_CROSS_counter);
 printf ("Total number of inserted hwrp's =%d out of %d possible.\n",
	 local_counter_of_inserted_hwrp,local_counter_group123456);



 // total hwrp connections:
 total_hwrp_switch_connections = 
   2 * (hwrp_H_counter + hwrp_V_counter) + // 2* because they are bidirectional.
   4 * hwrp_L_counter + // 4 because each hwrp_L_counter is two L's and each
                        // L is bidirectional.
   6 * hwrp_T_counter;  // 6 because there are 3-way switch-connections which
                        // are like a clique.
 total_efective_eliminated_due_to_hwrp_insertion =
   8 * hwrp_H_counter +
   8 * hwrp_L_counter +
   4 * hwrp_T_counter; 

 // Let's see how many switch-connections we have and how they determine the
 // total number of bufferes and pass transitors inside switch-blocks. 
 // i=0 is (i) usual switch-connection
 counter_conn_buffers[0] += 
   4 * (local_counter_group123456 - local_counter_of_inserted_hwrp) + 
   // 4 because each segment-ending entering the SB connects to other three
   // on the other three sides and there is buffer sharing.
   2 * hwrp_T_counter +
   // 2 because there is a usual switch-connection (vertical or horizontal)
   // accompanying every T hwrp in order not to have dangling segments.
   // These are bidirectional and have pass-transistors!
   2 * Num_Group7_type_connections;
   // Currently all Group7 connections are left as "usual". They are not
   // used for hwrp's insertion.
 counter_conn_pass_tran[0] += 
   12 * (local_counter_group123456 - local_counter_of_inserted_hwrp) + 
   2 * hwrp_T_counter +
   2 * Num_Group7_type_connections;

 // i=1 is (ii) hwrp switch-connection
 counter_conn_buffers[1] += 
   2 * (hwrp_H_counter + hwrp_V_counter) + // 2* because they are bidirectional.
   4 * hwrp_L_counter + // 4 because each hwrp_L_counter is two L's and each
                        // L is bidirectional.
   3 * hwrp_T_counter;  // 3 because there are 3-way switch-connections which
                        // are like a clique. Assumed buffer sharing.
 counter_conn_pass_tran[1] += 0; // hwrp's have no pass transistor. 




 // Debugging purposes.
 //dump_physical_array_switch_boxes("dbg_physical_array_sb.echo", nodes_per_chan); 
}//eof
// ======================================================================= //



enum e_hwrp_switch_types get_switch_type_HWRP (
	   short i, 
	   short j, 
	   short track,
	   int nodes_per_chan )
{
 // c.a:  Returns info about the switch type connection at the inquired
 // location and track number.  This info resides in 
 // "physical_array_switch_boxes", which was created and populated
 // already by generate_distributions_hwrp().
 // oBs:  I do not go thru the same messy procedure as get_switch_type()
 // because this routine is called after get_switch_type() and so I know
 // already that thre IS a connection.  The coordinate system of 
 // FPGA tiles is same as VPR's to keep things compatible.
  //
  // 10 20 30  41 42 43 44  51 52 53 54  60 70
  // H  V  C   L1 L2 L3 L4  T1 T2 T3 T4  NS NSAA

  enum e_hwrp_switch_types local_answer=NORMAL_SWITCH;


  if ( physical_array_switch_boxes[i][j][track] == 10 ) {
    // This is the Horizontal type of hwrp.
    local_answer = HORIZONTAL;
  }

  else if ( physical_array_switch_boxes[i][j][track] == 20 ) {
    local_answer = VERTICAL;
    //printf(" _%d",track);
  }

  /*---
  else if ( physical_array_switch_boxes[i][j][track] == 30 ) {
    //
    local_answer = CROSS;
  }
  ---*/

  else if ( physical_array_switch_boxes[i][j][track] == 41 ) {
    local_answer = L1;
  }
  else if ( physical_array_switch_boxes[i][j][track] == 42 ) {
    local_answer = L2;
  }
  else if ( physical_array_switch_boxes[i][j][track] == 43 ) {
    local_answer = L3;
  }
  else if ( physical_array_switch_boxes[i][j][track] == 44 ) {
    local_answer = L4;
  }

  else if ( physical_array_switch_boxes[i][j][track] == 51 ) {
    local_answer = T1;
  }
  else if ( physical_array_switch_boxes[i][j][track] == 52 ) {
    local_answer = T2;
  }
  else if ( physical_array_switch_boxes[i][j][track] == 53 ) {
    local_answer = T3;
  }
  else if ( physical_array_switch_boxes[i][j][track] == 54 ) {
    local_answer = T4;
  }

  /*---
  else if ( physical_array_switch_boxes[i][j][track] == -1 ) {
    //
    local_answer = NO_SWITCH_AT_ALL;
  }
  ---*/

  else {
    local_answer = NORMAL_SWITCH;
  }


  return local_answer;
}//eof
// ======================================================================= //


boolean is_it_ok_to_insert_hwrp_at_this_location(
		   int i,int j,int k,
		   int nodes_per_chan, 
		   t_seg_details *seg_details_x, 
		   t_seg_details *seg_details_y)
{
  // Returns true if at location i,j connection of four terminals
  // of this switch-block on track k would not lead to possibility
  // to have adjacent hwrp's, which would lead to connected trees
  // in the fpga and add extra capacitance of more dangling segments.
  boolean is_it_ok = TRUE;
  boolean top=TRUE, bottom=TRUE, right=TRUE, left=TRUE;
  int istart=0, iend=0, jstart=0, jend=0;


  if ( (i > 1) && (i < nx-1) && (j > 1) && (j < ny-1) ) {
    // location has neighbors on all four sides.
    jend = get_seg_end (seg_details_y, k, j+1, i, ny);//j+1 == jstart
    if (physical_array_switch_boxes[i][jend][k + 3*nodes_per_chan] <= 1)
      top = TRUE;
    else top = FALSE;

    jstart = get_closest_seg_start (seg_details_y, k, j, i);//
    if (physical_array_switch_boxes[i][jstart-1][k + 1*nodes_per_chan] <= 1)
      bottom = TRUE;
    else bottom = FALSE;

    iend = get_seg_end (seg_details_x, k, i+1, j, nx);
    if (physical_array_switch_boxes[iend][j][k] <= 1)
      right = TRUE;
    else right = FALSE;

    istart = get_closest_seg_start (seg_details_x, k, i, j);
    if (physical_array_switch_boxes[istart-1][j][k + 2*nodes_per_chan] <= 1)
      left = TRUE;
    else left = FALSE;

    if (top && bottom && right && left)  is_it_ok = TRUE;
    else                                 is_it_ok = FALSE;
    return is_it_ok;
  }


  else if ( (i > 1) && (i < nx-1) && (j == 1) ) {
    // location does not have neighbors to bottom.
    jend = get_seg_end (seg_details_y, k, j+1, i, ny);//j+1 == jstart
    if (physical_array_switch_boxes[i][jend][k + 3*nodes_per_chan] <= 1)
      top = TRUE;
    else top = FALSE;

    bottom = TRUE;// <--
      
    iend = get_seg_end (seg_details_x, k, i+1, j, nx);
    if (physical_array_switch_boxes[iend][j][k] <= 1)
      right = TRUE;
    else right = FALSE;

    istart = get_closest_seg_start (seg_details_x, k, i, j);
    if (physical_array_switch_boxes[istart-1][j][k + 2*nodes_per_chan] <= 1)
      left = TRUE;
    else left = FALSE;

    if (top && bottom && right && left)  is_it_ok = TRUE;
    else                                 is_it_ok = FALSE;
    return is_it_ok;
  }

  if ( (i > 1) && (i < nx-1) && (j == ny-1) ) {
    // location does not have neighbors to top.
    top = TRUE;// <--

    jstart = get_closest_seg_start (seg_details_y, k, j, i);//
    if (physical_array_switch_boxes[i][jstart-1][k + 1*nodes_per_chan] <= 1)
      bottom = TRUE;
    else bottom = FALSE;

    iend = get_seg_end (seg_details_x, k, i+1, j, nx);
    if (physical_array_switch_boxes[iend][j][k] <= 1)
      right = TRUE;
    else right = FALSE;

    istart = get_closest_seg_start (seg_details_x, k, i, j);
    if (physical_array_switch_boxes[istart-1][j][k + 2*nodes_per_chan] <= 1)
      left = TRUE;
    else left = FALSE;

    if (top && bottom && right && left)  is_it_ok = TRUE;
    else                                 is_it_ok = FALSE;
    return is_it_ok;
  }

  if ( (i == 1) && (j > 1) && (j < ny-1) ) {
    // location does not have neighbors to left.
    jend = get_seg_end (seg_details_y, k, j+1, i, ny);//j+1 == jstart
    if (physical_array_switch_boxes[i][jend][k + 3*nodes_per_chan] <= 1)
      top = TRUE;
    else top = FALSE;

    jstart = get_closest_seg_start (seg_details_y, k, j, i);//
    if (physical_array_switch_boxes[i][jstart-1][k + 1*nodes_per_chan] <= 1)
      bottom = TRUE;
    else bottom = FALSE;

    iend = get_seg_end (seg_details_x, k, i+1, j, nx);
    if (physical_array_switch_boxes[iend][j][k] <= 1)
      right = TRUE;
    else right = FALSE;

    left = TRUE; // <--

    if (top && bottom && right && left)  is_it_ok = TRUE;
    else                                 is_it_ok = FALSE;
    return is_it_ok;
  }

  if ( (i == nx-1) && (j > 1) && (j < ny-1) ) {
    // location does not have neighbors to right.
    jend = get_seg_end (seg_details_y, k, j+1, i, ny);//j+1 == jstart
    if (physical_array_switch_boxes[i][jend][k + 3*nodes_per_chan] <= 1)
      top = TRUE;
    else top = FALSE;

    jstart = get_closest_seg_start (seg_details_y, k, j, i);//
    if (physical_array_switch_boxes[i][jstart-1][k + 1*nodes_per_chan] <= 1)
      bottom = TRUE;
    else bottom = FALSE;

    right = TRUE;// <--

    istart = get_closest_seg_start (seg_details_x, k, i, j);
    if (physical_array_switch_boxes[istart-1][j][k + 2*nodes_per_chan] <= 1)
      left = TRUE;
    else left = FALSE;

    if (top && bottom && right && left)  is_it_ok = TRUE;
    else                                 is_it_ok = FALSE;
    return is_it_ok;
  }


  else if ( (i == 1) && (j == 1) ) {
    // location left-bottom corner.
    jend = get_seg_end (seg_details_y, k, j+1, i, ny);//j+1 == jstart
    if (physical_array_switch_boxes[i][jend][k + 3*nodes_per_chan] <= 1)
      top = TRUE;
    else top = FALSE;

    bottom = TRUE; // <--

    iend = get_seg_end (seg_details_x, k, i+1, j, nx);
    if (physical_array_switch_boxes[iend][j][k] <= 1)
      right = TRUE;
    else right = FALSE;

    left = TRUE;   // <--

    if (top && bottom && right && left)  is_it_ok = TRUE;
    else                                 is_it_ok = FALSE;
    return is_it_ok;
  }

  if ( (i == 1) && (j == ny-1) ) {
    // location left-top corner.
    top = TRUE;// <--
 
    jstart = get_closest_seg_start (seg_details_y, k, j, i);//
    if (physical_array_switch_boxes[i][jstart-1][k + 1*nodes_per_chan] <= 1)
      bottom = TRUE;
    else bottom = FALSE;

    iend = get_seg_end (seg_details_x, k, i+1, j, nx);
    if (physical_array_switch_boxes[iend][j][k] <= 1)
      right = TRUE;
    else right = FALSE;

    left = TRUE;// <--
 
    if (top && bottom && right && left)  is_it_ok = TRUE;
    else                                 is_it_ok = FALSE;
    return is_it_ok;
  }

  if ( (i == nx-1) && (j == 1) ) {
    // location right-bottom corner.
    jend = get_seg_end (seg_details_y, k, j+1, i, ny);//j+1 == jstart
    if (physical_array_switch_boxes[i][jend][k + 3*nodes_per_chan] <= 1)
      top = TRUE;
    else top = FALSE;

    bottom = TRUE; // <--
    right = TRUE;  // <--

    istart = get_closest_seg_start (seg_details_x, k, i, j);
    if (physical_array_switch_boxes[istart-1][j][k + 2*nodes_per_chan] <= 1)
      left = TRUE;
    else left = FALSE;

    if (top && bottom && right && left)  is_it_ok = TRUE;
    else                                 is_it_ok = FALSE;
    return is_it_ok;
  }

  if ( (i == nx-1) && (j == ny-1) ) {
    // location right-top corner.
    top = TRUE;   // <--

    jstart = get_closest_seg_start (seg_details_y, k, j, i);//
    if (physical_array_switch_boxes[i][jstart-1][k + 1*nodes_per_chan] <= 1)
      bottom = TRUE;
    else bottom = FALSE;

    right = TRUE; // <--

    istart = get_closest_seg_start (seg_details_x, k, i, j);
    if (physical_array_switch_boxes[istart-1][j][k + 2*nodes_per_chan] <= 1)
      left = TRUE;
    else left = FALSE;

    if (top && bottom && right && left)  is_it_ok = TRUE;
    else                                 is_it_ok = FALSE;
    return is_it_ok;
  }

  printf("Error in is_it_ok_to_insert_hwrp_at_this_location.\n");
  printf("Looking at invalid location: %d, %d, %d",i,j,k);
  exit(1);
}//eof
// ======================================================================= //


void dump_physical_array_switch_boxes(char *file_name, int nodes_per_chan) 
{
 // Debugging.
 int i=0, j=0, k=0;
 FILE *fp;

 fp = my_fopen (file_name, "wa", 0);

 fprintf (fp, "\n\n\n");
 for (i=1;i<nx;i++) { 
   for (j=1;j<ny;j++) {

     fprintf (fp, "[%d,%d]",i,j);

     for (k=0;k<nodes_per_chan;k++)
       fprintf (fp, " %d", physical_array_switch_boxes[i][j][k]);
     fprintf (fp, "  ");
     for (k=0;k<nodes_per_chan;k++)
       fprintf (fp, " %d", physical_array_switch_boxes[i][j][k + 1*nodes_per_chan]);
     fprintf (fp, "  ");
     for (k=0;k<nodes_per_chan;k++)
       fprintf (fp, " %d", physical_array_switch_boxes[i][j][k + 2*nodes_per_chan]);
     fprintf (fp, "  ");
     for (k=0;k<nodes_per_chan;k++)
       fprintf (fp, " %d", physical_array_switch_boxes[i][j][k + 3*nodes_per_chan]);
     fprintf (fp, "\n");

   }//for
 }//for


 fclose (fp);
}//eof
// ======================================================================= //
